# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import units

from manila.common import constants as const
from manila.share import configuration as conf
from manila.share.drivers.dell_emc.common.enas import utils
from manila.tests import fake_share


def query(func):
    def inner(*args, **kwargs):
        return (
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
            '<RequestPacket xmlns="http://www.emc.com/schemas/celerra/'
            'xml_api"><Request><Query>'
            + func(*args, **kwargs)
            + '</Query></Request></RequestPacket>'
        )

    return inner


def start_task(func):
    def inner(*args, **kwargs):
        return (
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
            '<RequestPacket xmlns="http://www.emc.com/schemas/celerra/'
            'xml_api"><Request><StartTask timeout="300">'
            + func(*args, **kwargs)
            + '</StartTask></Request></RequestPacket>')

    return inner


def response(func):
    def inner(*args, **kwargs):
        return (
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
            '<ResponsePacket xmlns="http://www.emc.com/schemas/celerra/'
            'xml_api"><Response>'
            + func(*args, **kwargs)
            + '</Response></ResponsePacket>'
        ).encode()

    return inner


class FakeData(object):
    # Share information
    share_id = '7cf7c200_d3af_4e05_b87e_9167c95df4f9'
    host = 'HostA@BackendB#fake_pool_name'
    share_name = share_id
    share_size = 10
    new_size = 20
    src_share_name = '7cf7c200_d3af_4e05_b87e_9167c95df4f0'

    # Snapshot information
    snapshot_name = 'de4c9050-e2f9-4ce1-ade4-5ed0c9f26451'
    src_snap_name = 'de4c9050-e2f9-4ce1-ade4-5ed0c9f26452'
    snapshot_id = 'fake_snap_id'
    snapshot_size = 10 * units.Ki

    # Share network information
    share_network_id = 'c5b3a865-56d0-4d88-abe5-879965e099c9'
    cidr = '192.168.1.0/24'
    cidr_v6 = 'fdf8:f53b:82e1::/64'
    segmentation_id = 100
    network_allocations_id1 = '132dbb10-9a36-46f2-8d89-3d909830c356'
    network_allocations_id2 = '7eabdeed-bad2-46ea-bd0f-a33884c869e0'
    network_allocations_id3 = '98c9e490-a842-4e59-b59a-a6042069d35b'
    network_allocations_id4 = '6319a917-ab95-4b65-a498-773ae33c5550'
    network_allocations_ip1 = '192.168.1.1'
    network_allocations_ip2 = '192.168.1.2'
    network_allocations_ip3 = 'fdf8:f53b:82e1::1'
    network_allocations_ip4 = 'fdf8:f53b:82e1::2'

    network_allocations_ip_version1 = 4
    network_allocations_ip_version2 = 4
    network_allocations_ip_version3 = 6
    network_allocations_ip_version4 = 6
    domain_name = 'fake_domain'
    domain_user = 'administrator'
    domain_password = 'password'
    dns_ip_address = '192.168.1.200'
    dns_ipv6_address = 'fdf8:f53b:82e1::f'

    # Share server information
    share_server_id = '56aafd02-4d44-43d7-b784-57fc88167224'

    # Filesystem information
    filesystem_name = share_name
    filesystem_id = 'fake_filesystem_id'
    filesystem_size = 10 * units.Ki
    filesystem_new_size = 20 * units.Ki

    # Mountpoint information
    path = '/' + share_name

    # Mover information
    mover_name = 'server_2'
    mover_id = 'fake_mover_id'
    interface_name1 = network_allocations_id1[-12:]
    interface_name2 = network_allocations_id2[-12:]
    interface_name3 = network_allocations_id3[-12:]
    interface_name4 = network_allocations_id4[-12:]
    long_interface_name = network_allocations_id1
    net_mask = '255.255.255.0'
    net_mask_v6 = 64
    device_name = 'cge-1-0'
    interconnect_id = '2001'

    # VDM information
    vdm_name = share_server_id
    vdm_id = 'fake_vdm_id'

    # Pool information
    pool_name = 'fake_pool_name'
    pool_id = 'fake_pool_id'
    pool_used_size = 20480
    pool_total_size = 511999

    # NFS share access information
    rw_hosts = ['192.168.1.1', '192.168.1.2']
    ro_hosts = ['192.168.1.3', '192.168.1.4']
    nfs_host_ip = '192.168.1.5'
    rw_hosts_ipv6 = ['fdf8:f53b:82e1::1', 'fdf8:f53b:82e1::2']
    ro_hosts_ipv6 = ['fdf8:f53b:82e1::3', 'fdf8:f53b:82e1::4']
    nfs_host_ipv6 = 'fdf8:f53b:82e1::5'

    fake_output = ''

    fake_error_msg = 'fake error message'

    emc_share_backend = 'vnx'
    vmax_share_backend = 'vmax'
    emc_nas_server = '192.168.1.20'
    emc_nas_login = 'fakename'
    emc_nas_password = 'fakepassword'
    share_backend_name = 'EMC_NAS_Storage'

    cifs_access = """
        1478607389: SMB:11: Unix user 'Guest' UID=32769
        1478607389: SMB:10: FindUserUid:Access_Password 'Guest',1=0x8001 T=0
        1478607389: SHARE: 6: ALLOWED:fullcontrol:S-1-5-15-3399d125-6dcdf5f4
        1478607389: SMB:11: Unix user 'Administrator' UID=32768
        1478607389: SMB:10: FindUserUid:Access_Password 'Administrator',
        1478607389: SHARE: 6: ALLOWED:fullcontrol:S-1-5-15-3399d125
        """


class StorageObjectTestData(object):
    def __init__(self):
        self.share_name = FakeData.share_name

        self.filesystem_name = FakeData.filesystem_name
        self.filesystem_id = FakeData.filesystem_id
        self.filesystem_size = 10 * units.Ki
        self.filesystem_new_size = 20 * units.Ki

        self.path = FakeData.path

        self.snapshot_name = FakeData.snapshot_name
        self.snapshot_id = FakeData.snapshot_id
        self.snapshot_size = 10 * units.Ki

        self.src_snap_name = FakeData.src_snap_name
        self.src_fileystems_name = FakeData.src_share_name

        self.mover_name = FakeData.mover_name
        self.mover_id = FakeData.mover_id
        self.vdm_name = FakeData.vdm_name
        self.vdm_id = FakeData.vdm_id

        self.pool_name = FakeData.pool_name
        self.pool_id = FakeData.pool_id
        self.pool_used_size = FakeData.pool_used_size
        self.pool_total_size = FakeData.pool_total_size

        self.interface_name1 = FakeData.interface_name1
        self.interface_name2 = FakeData.interface_name2
        self.interface_name3 = FakeData.interface_name3
        self.interface_name4 = FakeData.interface_name4
        self.long_interface_name = FakeData.long_interface_name
        self.ip_address1 = FakeData.network_allocations_ip1
        self.ip_address2 = FakeData.network_allocations_ip2
        self.ip_address3 = FakeData.network_allocations_ip3
        self.ip_address4 = FakeData.network_allocations_ip4
        self.net_mask = FakeData.net_mask
        self.net_mask_v6 = FakeData.net_mask_v6
        self.vlan_id = FakeData.segmentation_id

        self.cifs_server_name = FakeData.vdm_name

        self.domain_name = FakeData.domain_name
        self.domain_user = FakeData.domain_user
        self.domain_password = FakeData.domain_password
        self.dns_ip_address = FakeData.dns_ip_address

        self.device_name = FakeData.device_name

        self.interconnect_id = FakeData.interconnect_id

        self.rw_hosts = FakeData.rw_hosts
        self.ro_hosts = FakeData.ro_hosts
        self.nfs_host_ip = FakeData.nfs_host_ip

        self.rw_hosts_ipv6 = FakeData.rw_hosts_ipv6
        self.ro_hosts_ipv6 = FakeData.ro_hosts_ipv6
        self.nfs_host_ipv6 = FakeData.nfs_host_ipv6

        self.fake_output = FakeData.fake_output

    @response
    def resp_get_error(self):
        return (
            '<QueryStatus maxSeverity="error">'
            '<Problem messageCode="18522112101" facility="Generic" '
            'component="API" message="Fake message." severity="error">'
            '<Description>Fake description.</Description>'
            '<Action>Fake action.</Action>'
            '<Diagnostics>Fake diagnostics.</Diagnostics>'
            '</Problem>'
            '<Problem messageCode="18522112101" facility="Generic" '
            'component="API" message="Fake message." severity="error">'
            '<Description>Fake description.</Description>'
            '<Action>Fake action.</Action>'
            '<Diagnostics>Fake diagnostics.</Diagnostics>'
            '</Problem>'
            '</QueryStatus> '
        )

    @response
    def resp_get_without_value(self):
        return (
            '<QueryStatus maxSeverity="ok"/>'
        )

    @response
    def resp_task_succeed(self):
        return (
            '<TaskResponse taskId="123">'
            '<Status maxSeverity="ok"/>'
            '</TaskResponse>'
        )

    @response
    def resp_task_error(self):
        return (
            '<TaskResponse taskId="123">'
            '<Status maxSeverity="error"/>'
            '</TaskResponse>'
        )

    @response
    def resp_invalid_mover_id(self):
        return (
            '<Fault maxSeverity="error">'
            '<Problem messageCode="14227341323" facility="Prevalidator" '
            'component="API" message="Mover with id=100 not found." '
            'severity="error">'
            '<Description>The Mover ID supplied with the request is invalid.'
            '</Description>'
            '<Action>Refer to the XML API v2 schema/documentation and correct '
            'your user program logic.</Action>'
            '<Diagnostics> Exception tag: 14fb692e556 Exception '
            'message: com.emc.nas.ccmd.common.MessageInstanceImpl@5004000d '
            '</Diagnostics>'
            '</Problem>'
            '</Fault> '
        )

    @response
    def resp_need_retry(self):
        return ('<TaskResponse taskId="915525">'
                '<Status maxSeverity = "error">'
                '<Problem messageCode = "13421840537" component = "fake"'
                '    message = "unable to acquire lock(s), try later"'
                '    severity = "error" >'
                '<Description> fake desp. </Description>'
                '<Action>fake action </Action>'
                '</Problem></Status></TaskResponse>')

    @start_task
    def req_fake_start_task(self):
        return '<StartFake name="foo"></StartFake>'


class FileSystemTestData(StorageObjectTestData):
    def __init__(self):
        super(FileSystemTestData, self).__init__()

    @start_task
    def req_create_on_vdm(self):
        return (
            '<NewFileSystem name="%(name)s">'
            '<Vdm vdm="%(id)s"/>'
            '<StoragePool mayContainSlices="true" pool="%(pool_id)s" '
            'size="%(size)s"/>'
            '</NewFileSystem>'
            % {'name': self.filesystem_name,
               'id': self.vdm_id,
               'pool_id': self.pool_id,
               'size': self.filesystem_size}
        )

    @start_task
    def req_create_on_mover(self):
        return (
            '<NewFileSystem name="%(name)s">'
            '<Mover mover="%(id)s"/>'
            '<StoragePool mayContainSlices="true" pool="%(pool_id)s" '
            'size="%(size)s"/>'
            '</NewFileSystem>'
            % {'name': self.filesystem_name,
               'id': self.mover_id,
               'pool_id': self.pool_id,
               'size': self.filesystem_size}
        )

    @response
    def resp_create_but_already_exist(self):
        return (
            ' <TaskResponse taskId="31362">'
            '<Status maxSeverity="error">'
            '<Problem messageCode="13691191325" component="APL" '
            'message="A file system with the name fake_filesystem '
            'already exists." severity="error">'
            '<Description></Description>'
            '<Action></Action>'
            '</Problem>'
            '</Status>'
            '</TaskResponse> '
        )

    @start_task
    def req_delete(self):
        return (
            '<DeleteFileSystem fileSystem="%(id)s"/>' %
            {'id': self.filesystem_id}
        )

    @response
    def resp_delete_but_failed(self):
        return (
            '<Fault maxSeverity="error">'
            '<Problem messageCode="14227341326" facility="Prevalidator" '
            'component="API" message="File system with id=77777 not found." '
            'severity="error">'
            '<Description>The file system ID supplied with the request is '
            'invalid.</Description>'
            '<Action>Refer to the XML API v2 schema/documentation and correct '
            'your user program logic.</Action>'
            '<Diagnostics> Exception tag: 14fb6b6a7b8 Exception '
            'message: com.emc.nas.ccmd.common.MessageInstanceImpl@5004000e '
            '</Diagnostics>'
            '</Problem>'
            '</Fault> '
        )

    @start_task
    def req_extend(self):
        return (
            '<ExtendFileSystem fileSystem="%(id)s">'
            '<StoragePool pool="%(pool_id)s" size="%(size)d"/>'
            '</ExtendFileSystem>' %
            {'id': self.filesystem_id,
             'pool_id': self.pool_id,
             'size': self.filesystem_new_size - self.filesystem_size}
        )

    @response
    def resp_extend_but_error(self):
        return (
            '<Fault maxSeverity="error">'
            '<Problem messageCode="14227341325" facility="Prevalidator" '
            'component="API" message="Fake message.">'
            '<Description>Fake description.</Description>'
            '<Action>Fake action.</Action>'
            '<Diagnostics> Fake diagnostics.</Diagnostics>'
            '</Problem>'
            '</Fault> '
        )

    @query
    def req_get(self):
        return (
            '<FileSystemQueryParams>'
            '<AspectSelection fileSystemCapacityInfos="true" '
            'fileSystems="true"/>'
            '<Alias name="%(name)s"/>'
            '</FileSystemQueryParams>' %
            {'name': self.filesystem_name}
        )

    @response
    def resp_get_succeed(self):
        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<FileSystem name="%(name)s" type="uxfs" volume="107" '
            'storagePools="%(pool_id)s" storages="1" containsSlices="true" '
            'internalUse="false" dataServicePolicies="Thin=No,Compressed=No,'
            'Mirrored=No,Tiering policy=Auto-Tier/Highest Available Tier" '
            'fileSystem="%(id)s">'
            '<ProductionFileSystemData cwormState="off"/>'
            '</FileSystem>'
            '<FileSystemCapacityInfo volumeSize="%(size)s" '
            'fileSystem="%(id)s"/>' %
            {'name': self.filesystem_name,
             'id': self.filesystem_id,
             'size': self.filesystem_size,
             'pool_id': self.pool_id}
        )

    @response
    def resp_get_but_miss_property(self):
        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<FileSystem name="%(name)s" type="uxfs" volume="107" '
            'storagePools="%(pool_id)s" storages="1" containsSlices="true" '
            'internalUse="false" '
            'fileSystem="%(id)s">'
            '<ProductionFileSystemData cwormState="off"/>'
            '</FileSystem>'
            '<FileSystemCapacityInfo volumeSize="%(size)s" '
            'fileSystem="%(id)s"/>' %
            {'name': self.filesystem_name,
             'id': self.filesystem_id,
             'size': self.filesystem_size,
             'pool_id': self.pool_id}
        )

    @response
    def resp_get_but_not_found(self):
        return (
            '<QueryStatus maxSeverity="warning">'
            '<Problem messageCode="18522112101" facility="Generic" '
            'component="API" message="The query may be incomplete or '
            'requested object not found." severity="warning">'
            '<Description>The query may be incomplete because some of the '
            'Celerra components are unavailable or do not exist. Another '
            'reason may be application error. </Description>'
            '<Action>If the entire Celerra is functioning correctly, '
            'check your client application logic. </Action>'
            '<Diagnostics>File system not found.</Diagnostics>'
            '</Problem>'
            '<Problem messageCode="18522112101" facility="Generic" '
            'component="API" message="The query may be incomplete or '
            'requested object not found." severity="warning">'
            '<Description>The query may be incomplete because some of the '
            'Celerra components are unavailable or do not exist. Another '
            'reason may be application error.</Description>'
            '<Action>If the entire Celerra is functioning correctly, '
            'check your client application logic.</Action>'
            '<Diagnostics>Migration file system not found.</Diagnostics>'
            '</Problem>'
            '</QueryStatus> '
        )

    def cmd_create_from_ckpt(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
            '-name', self.filesystem_name,
            '-type', 'uxfs',
            '-create',
            'samesize=' + self.src_fileystems_name,
            'pool=' + self.pool_name,
            'storage=SINGLE',
            'worm=off',
            '-thin', 'no',
            '-option', 'slice=y',
        ]

    def cmd_copy_ckpt(self):
        session_name = self.filesystem_name + ':' + self.src_snap_name

        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_copy',
            '-name', session_name[0:63],
            '-source', '-ckpt', self.src_snap_name,
            '-destination', '-fs', self.filesystem_name,
            '-interconnect', "id=" + self.interconnect_id,
            '-overwrite_destination',
            '-full_copy',
        ]

    output_copy_ckpt = "OK"
    error_copy_ckpt = "ERROR"

    def cmd_nas_fs_info(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
            '-info', self.filesystem_name,
        ]

    def output_info(self):
        return (
            """output = id        = 515
                name      = %(share_name)s
                acl       = 0
                in_use    = True
                type      = uxfs
                worm      = off
                volume    = v993
                deduplication   = Off
                thin_storage    = True
                tiering_policy  = Auto-Tier/Optimize Pool
                compressed= False
                mirrored  = False
                ckpts     = %(ckpt)s
                stor_devs = FNM00124500890-004B
                disks     = d7
                 disk=d7    fakeinfo""" %
            {'share_name': self.filesystem_name,
             'ckpt': self.snapshot_name})

    def cmd_delete(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
            '-delete', self.snapshot_name,
            '-Force',
        ]


class SnapshotTestData(StorageObjectTestData):
    def __init__(self):
        super(SnapshotTestData, self).__init__()

    @start_task
    def req_create(self):
        return (
            '<NewCheckpoint checkpointOf="%(fsid)s" '
            'name="%(name)s"><SpaceAllocationMethod>'
            '<StoragePool pool="%(pool_id)s"/></SpaceAllocationMethod>'
            '</NewCheckpoint>'
            % {'fsid': self.filesystem_id,
               'name': self.snapshot_name,
               'pool_id': self.pool_id}
        )

    @start_task
    def req_create_with_size(self):
        return (
            '<NewCheckpoint checkpointOf="%(fsid)s" '
            'name="%(name)s"><SpaceAllocationMethod>'
            '<StoragePool pool="%(pool_id)s" size="%(size)s"/>'
            '</SpaceAllocationMethod>'
            '</NewCheckpoint>'
            % {'fsid': self.filesystem_id,
               'name': self.snapshot_name,
               'pool_id': self.pool_id,
               'size': self.snapshot_size}
        )

    @response
    def resp_create_but_already_exist(self):
        return (
            '<Status maxSeverity="error">'
            '<Problem messageCode="13690535947" component="APL" '
            'message="snap_0 is already in use." severity="error">'
            '<Description></Description>'
            '<Action></Action>'
            '</Problem>'
            '</Status>'
        )

    @query
    def req_get(self):
        return (
            '<CheckpointQueryParams><Alias name="%(name)s"/>'
            '</CheckpointQueryParams>'
            % {'name': self.snapshot_name}
        )

    @response
    def resp_get_succeed(self):
        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<Checkpoint checkpointOf="%(fs_id)s" name="%(name)s_replica1"'
            ' state="active" time="1405428355" fileSystemSize="0"'
            ' checkpoint="%(snap_id)s"/>'
            % {'name': self.snapshot_name,
               'fs_id': self.filesystem_id,
               'snap_id': self.snapshot_id}
        )

    @start_task
    def req_delete(self):
        return (
            '<DeleteCheckpoint checkpoint="%(id)s"/>' %
            {'id': self.snapshot_id}
        )


class MountPointTestData(StorageObjectTestData):
    def __init__(self):
        super(MountPointTestData, self).__init__()

    @start_task
    def req_create(self, mover_id, is_vdm=True):
        return (
            '<NewMount path="%(path)s" fileSystem="%(fs_id)s">'
            '<MoverOrVdm mover="%(mover_id)s" moverIdIsVdm="%(is_vdm)s"/>'
            '</NewMount>' %
            {'path': self.path,
             'fs_id': self.filesystem_id,
             'mover_id': mover_id,
             'is_vdm': 'true' if is_vdm else 'false'}
        )

    @response
    def resp_create_but_already_exist(self):
        return (
            '<TaskResponse taskId="31428">'
            '<Status maxSeverity="error">'
            '<Problem messageCode="13690601492" component="APL" '
            'message="Mount already exists" severity="error">'
            '<Description></Description> <Action></Action>'
            '</Problem>'
            '</Status>'
            '</TaskResponse> ')

    @start_task
    def req_delete(self, mover_id, is_vdm=True):
        return (
            '<DeleteMount path="%(path)s" mover="%(mover_id)s" '
            'moverIdIsVdm="%(is_vdm)s"/>' %
            {'path': self.path,
             'mover_id': mover_id,
             'is_vdm': 'true' if is_vdm else 'false'}
        )

    @response
    def resp_delete_but_nonexistent(self):
        return (
            '<TaskResponse taskId="31401">'
            '<Status maxSeverity="error"> <Problem messageCode="13690601492" '
            'component="APL" message="/fake_filesystem : No such path or '
            'invalid operation." severity="error">'
            '<Description></Description> <Action></Action>'
            '</Problem>'
            '</Status>'
            '</TaskResponse> '
        )

    @query
    def req_get(self, mover_id, is_vdm=True):
        return (
            '<MountQueryParams><MoverOrVdm mover="%(mover_id)s" '
            'moverIdIsVdm="%(is_vdm)s"/></MountQueryParams>' %
            {'mover_id': mover_id,
             'is_vdm': 'true' if is_vdm else 'false'}
        )

    @response
    def resp_get_succeed(self, mover_id, is_vdm=True):
        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<Mount fileSystem="%(fsID)s" disabled="false"'
            ' ntCredential="false" path="%(path)s" mover="%(mover_id)s"'
            ' moverIdIsVdm="%(is_vdm)s">'
            '<NfsOptions ro="false" virusScan="true"'
            ' prefetch="true" uncached="false"/>'
            '<CifsOptions cifsSyncwrite="false" notify="true"'
            ' triggerLevel="512" notifyOnAccess="false"'
            ' notifyOnWrite="false" oplock="true" accessPolicy="NATIVE"'
            ' lockingPolicy="nolock"/></Mount>'
            % {'path': self.path,
               'fsID': self.filesystem_id,
               'mover_id': mover_id,
               'is_vdm': 'true' if is_vdm else 'false'}
        )

    def cmd_server_mount(self, mode):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/server_mount', self.vdm_name,
            '-option', mode,
            self.filesystem_name,
            self.path,
        ]

    def cmd_server_umount(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/server_umount', self.vdm_name,
            '-perm', self.snapshot_name,
        ]


class VDMTestData(StorageObjectTestData):
    def __init__(self):
        super(VDMTestData, self).__init__()

    @start_task
    def req_create(self):
        return (
            '<NewVdm mover="%(mover_id)s" name="%(vdm_name)s"/>' %
            {'mover_id': self.mover_id, 'vdm_name': self.vdm_name}
        )

    @response
    def resp_create_but_already_exist(self):
        return (
            '<TaskResponse taskId="32551">'
            '<Status maxSeverity="error">'
            '<Problem messageCode="13421840550" component="CS_CORE" '
            'message="VDM_01 : an entry with this name already exists" '
            'severity="error">'
            '<Description>Duplicate name specified</Description>'
            '<Action>Specify a unqiue name</Action>'
            '</Problem>'
            '<Problem messageCode="13421840550" component="CS_CORE" '
            'message="VDM_01 : an entry with this name already exists" '
            'severity="error">'
            '<Description>Duplicate name specified</Description>'
            '<Action>Specify a unqiue name</Action>'
            '</Problem>'
            '</Status>'
            '</TaskResponse> '
        )

    @query
    def req_get(self):
        return '<VdmQueryParams/>'

    @response
    def resp_get_succeed(self, name=None, interface1=None, interface2=None):
        if name is None:
            name = self.vdm_name

        if interface1 is None:
            interface1 = self.interface_name1

        if interface2 is None:
            interface2 = self.interface_name2

        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<Vdm name="%(vdm_name)s" state="loaded" mover="%(mover_id)s" '
            'rootFileSystem="396" vdm="%(vdm_id)s">'
            '<Status maxSeverity="ok"/>'
            '<Interfaces> <li>%(interface1)s</li> <li>%(interface2)s</li>'
            '</Interfaces> </Vdm>' %
            {'vdm_name': name,
             'vdm_id': self.vdm_id,
             'mover_id': self.mover_id,
             'interface1': interface1,
             'interface2': interface2}
        )

    @response
    def resp_get_but_not_found(self):
        return (
            '<QueryStatus maxSeverity="ok"/>'
        )

    @start_task
    def req_delete(self):
        return '<DeleteVdm vdm="%(vdmid)s"/>' % {'vdmid': self.vdm_id}

    def cmd_attach_nfs_interface(self, interface=None):
        if interface is None:
            interface = self.interface_name2

        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_server',
            '-vdm', self.vdm_name,
            '-attach', interface,
        ]

    def cmd_detach_nfs_interface(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_server',
            '-vdm', self.vdm_name,
            '-detach', self.interface_name2,
        ]

    def cmd_get_interfaces(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_server',
            '-i',
            '-vdm', self.vdm_name,
        ]

    def output_get_interfaces_vdm(self,
                                  cifs_interface=FakeData.interface_name1,
                                  nfs_interface=FakeData.interface_name2):
        return (
            """id        = %(vdmid)s
            name      = %(name)s
            acl       = 0
            type      = vdm
            server    = server_2
            rootfs    = root_fs_vdm_vdm-fakeid
            I18N mode = UNICODE
            mountedfs =
            member_of =
            status    :
              defined = enabled
               actual = loaded, active
            Interfaces to services mapping:
             interface=%(nfs_if_name)s :vdm
             interface=%(cifs_if_name)s :cifs""" %
            {'vdmid': self.vdm_id,
             'name': self.vdm_name,
             'nfs_if_name': nfs_interface,
             'cifs_if_name': cifs_interface}
        )

    def output_get_interfaces_nfs(self,
                                  cifs_interface=FakeData.interface_name1,
                                  nfs_interface=FakeData.interface_name2):
        return (
            """id        = %(vdmid)s
            name      = %(name)s
            acl       = 0
            type      = vdm
            server    = server_2
            rootfs    = root_fs_vdm_vdm-fakeid
            I18N mode = UNICODE
            mountedfs =
            member_of =
            status    :
              defined = enabled
               actual = loaded, active
            Interfaces to services mapping:
             interface=%(nfs_if_name)s :nfs
             interface=%(cifs_if_name)s :cifs""" %
            {'vdmid': self.vdm_id,
             'name': self.vdm_name,
             'nfs_if_name': nfs_interface,
             'cifs_if_name': cifs_interface}
        )


class PoolTestData(StorageObjectTestData):
    def __init__(self):
        super(PoolTestData, self).__init__()

    @query
    def req_get(self):
        return (
            '<StoragePoolQueryParams/>'
        )

    @response
    def resp_get_succeed(self, name=None, id=None):
        if not name:
            name = self.pool_name
        if not id:
            id = self.pool_id
        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<StoragePool movers="1 2" memberVolumes="98" storageSystems="1" '
            'name="fake" '
            'description="Mapped Pool Pool_2 on APM00152904560" '
            'mayContainSlicesDefault="true" diskType="Capacity" '
            'size="511999" usedSize="20480" autoSize="511999" '
            'virtualProvisioning="false" isHomogeneous="true" '
            'dataServicePolicies="Thin=No,Compressed=No,Mirrored=No,Tiering '
            'policy=Auto-Tier/Highest Available Tier" templatePool="59" '
            'stripeCount="5" stripeSize="256" pool="9">'
            '<SystemStoragePoolData dynamic="true" greedy="true" '
            'potentialAdditionalSize="0" isBackendPool="true"/>'
            '</StoragePool>'
            '<StoragePool movers="1 2" memberVolumes="98" storageSystems="1" '
            'name="%(name)s" '
            'description="Mapped Pool Pool_2 on APM00152904560" '
            'mayContainSlicesDefault="true" diskType="Capacity" '
            'size="411999" usedSize="%(pool_used_size)s" '
            'autoSize="%(pool_total_size)s" virtualProvisioning="false" '
            'isHomogeneous="true" '
            'dataServicePolicies="Thin=No,Compressed=No,Mirrored=No,Tiering '
            'policy=Auto-Tier/Highest Available Tier" templatePool="59" '
            'stripeCount="5" stripeSize="256" pool="%(id)s">'
            '<SystemStoragePoolData dynamic="true" greedy="true" '
            'potentialAdditionalSize="0" isBackendPool="true"/>'
            '</StoragePool>' %
            {'name': name,
             'id': id,
             'pool_used_size': self.pool_used_size,
             'pool_total_size': self.pool_total_size}
        )


class MoverTestData(StorageObjectTestData):
    def __init__(self):
        super(MoverTestData, self).__init__()

    @query
    def req_get_ref(self):
        return (
            '<MoverQueryParams>'
            '<AspectSelection movers="true"/>'
            '</MoverQueryParams>'
        )

    @response
    def resp_get_ref_succeed(self, name=None):
        if not name:
            name = self.mover_name
        return (
            '<QueryStatus maxSeverity="info">'
            '<Problem messageCode="18522112101" facility="Generic" '
            'component="API" message="The query may be incomplete or '
            'requested object not found." severity="warning">'
            '<Description>The query may be incomplete because some of the '
            'Celerra components are unavailable or do not exist. Another '
            'reason may be application error.</Description>'
            '<Action>If the entire Celerra is functioning correctly, '
            'check your client application logic.</Action>'
            '<Diagnostics>Standby Data Mover server_2.faulted.server_3 is '
            'out of service.</Diagnostics>'
            '</Problem>'
            '</QueryStatus>'
            '<Mover name="%(name)s" host="1" role="primary" standbys="2" '
            'i18NMode="UNICODE" failoverPolicy="auto" '
            'ntpServers="192.168.1.82" mover="%(id)s"/>'
            '<Mover name="server_3" host="2" role="standby" standbyFors="1" '
            'i18NMode="ASCII" failoverPolicy="none" mover="2"/>' %
            {'name': name, 'id': self.mover_id}
        )

    @query
    def req_get(self):
        return (
            '<MoverQueryParams mover="%(id)s">'
            '<AspectSelection moverInterfaces="true" moverStatuses="true" '
            'movers="true" moverNisDomains="true" moverNetworkDevices="true" '
            'moverDnsDomains="true" moverRoutes="true" '
            'moverDeduplicationSettings="true"/>'
            '</MoverQueryParams>' %
            {'id': self.mover_id}
        )

    @response
    def resp_get_succeed(self, name=None):
        if not name:
            name = self.mover_name
        return (
            '<QueryStatus maxSeverity="ok"/><Mover name="%(name)s" '
            'host="1" role="primary" i18NMode="UNICODE" failoverPolicy="none"'
            ' ntpServers="192.168.1.82" mover="%(id)s"/>'
            '<MoverStatus version="T8.1.3.34944" csTime="1406795150" '
            'clock="140681" timezone="GMT-5" uptime="85096" '
            'mover="%(id)s"><Status maxSeverity="ok"/>'
            '</MoverStatus>'
            '<MoverDnsDomain servers="192.168.1.82" protocol="udp" '
            'mover="%(id)s" name="win2012.openstack"/>'
            '<MoverInterface name="%(long_interface_name)s" device="mge0" '
            'ipVersion="IPv4" netMask="255.255.255.0" '
            'broadcastAddr="128.221.252.255" '
            'macAddr="0:60:16:53:cc:87" mtu="1500" up="true" vlanid="0" '
            'mover="%(id)s" ipAddress="128.221.252.1"/>'
            '<MoverInterface name="%(interface_name1)s" device="cge-2-0" '
            'ipVersion="IPv4" netMask="255.255.255.0" '
            'broadcastAddr="128.221.252.255" macAddr="0:60:16:53:cc:87" '
            'mtu="1500" up="true" vlanid="0" mover="%(id)s" '
            'ipAddress="128.221.252.2"/>'
            '<MoverInterface name="%(interface_name2)s" device="cge-2-1" '
            'ipVersion="IPv4" netMask="255.255.255.0" '
            'broadcastAddr="128.221.252.255" macAddr="0:60:16:53:cc:87" '
            'mtu="1500" up="true" vlanid="0" mover="%(id)s" '
            'ipAddress="128.221.252.3"/>'
            '<MoverRoute destination="0.0.0.0" interface="192.168.1.178" '
            'ipVersion="IPv4" netMask="0.0.0.0" '
            'gateway="192.168.1.217" mover="%(id)s"/>'
            '<LogicalNetworkDevice speed="auto" interfaces="192.168.1.136" '
            'type="physical-ethernet" mover="%(id)s" name="cge-2-0"/>'
            % {'id': self.mover_id,
               'name': name,
               'long_interface_name': self.long_interface_name[:31],
               'interface_name1': self.interface_name1,
               'interface_name2': self.interface_name2}
        )

    @start_task
    def req_create_interface(self,
                             if_name=FakeData.interface_name1,
                             ip=FakeData.network_allocations_ip1):
        return (
            '<NewMoverInterface name="%(if_name)s" vlanid="%(vlan)s" '
            'netMask="%(net_mask)s" device="%(device_name)s" '
            'mover="%(mover_id)s" ipAddress="%(ip)s"/>'
            % {'if_name': if_name,
               'vlan': self.vlan_id,
               'ip': ip,
               'mover_id': self.mover_id,
               'device_name': self.device_name,
               'net_mask': self.net_mask}
        )

    @start_task
    def req_create_interface_with_ipv6(self,
                                       if_name=FakeData.interface_name3,
                                       ip=FakeData.network_allocations_ip3):
        return (
            '<NewMoverInterface name="%(if_name)s" vlanid="%(vlan)s" '
            'ipVersion="IPv6" netMask="%(net_mask)s" '
            'device="%(device_name)s" '
            'mover="%(mover_id)s" ipAddress="%(ip)s"/>'
            % {'if_name': if_name,
               'vlan': self.vlan_id,
               'ip': ip,
               'mover_id': self.mover_id,
               'device_name': self.device_name,
               'net_mask': self.net_mask_v6}
        )

    @response
    def resp_create_interface_but_name_already_exist(self):
        return (
            '<Status maxSeverity="error">'
            '<Problem messageCode="13421840550" component="CS_CORE" '
            'message="%(interface_name)s : an entry with this name already '
            'exists" severity="error">'
            '<Description>Duplicate name specified</Description>'
            '<Action>Specify a unqiue name</Action>'
            '</Problem>'
            '</Status>' % {'interface_name': self.interface_name1}
        )

    @response
    def resp_create_interface_but_ip_already_exist(self):
        return (
            '<Status maxSeverity="error">'
            '<Problem messageCode="13691781136" component="CS_CORE" '
            'message="Interface %(ip)s already exists."  severity="error">'
            '<Description></Description><Action></Action>'
            '</Problem>'
            '</Status>' % {'ip': self.ip_address1}
        )

    @response
    def resp_create_interface_with_conflicted_vlan_id(self):
        return (
            '<Status maxSeverity="error">'
            '<Problem messageCode="13421850371" component="CS_CORE" '
            'message="160: Invalid VLAN change. Other interfaces on this '
            'subnet are in a different VLAN." severity="error">'
            '<Description>The operation cannot complete because other '
            'interfaces on the same subnet are in a different VLAN. '
            'The Data Mover requires all interfaces in the same subnet '
            'to be in the same VLAN.</Description>'
            '<Action>Specify a VLAN to match other interfaces in the same '
            'subnet. To move multiple interfaces to a different VLAN, '
            'first set the VLAN id on each interface to 0, '
            'and then set their VLAN id\'s to the new VLAN number.</Action>'
            '</Problem>'
            '</Status>'
        )

    @start_task
    def req_delete_interface(self, ip=FakeData.network_allocations_ip1):

        return (
            '<DeleteMoverInterface mover="%(mover_id)s" '
            'ipAddress="%(ip)s"/>' %
            {'ip': ip,
             'mover_id': self.mover_id, }
        )

    @response
    def resp_delete_interface_but_nonexistent(self):
        return (
            '<Status maxSeverity="error">'
            '<Problem messageCode="13691781134" component="APL" '
            'message="Device 192.168.237.100 does not exist." '
            'severity="error">'
            '<Description></Description>'
            '<Action></Action>'
            '</Problem>'
            '</Status>'
        )

    def cmd_get_interconnect_id(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_cel',
            '-interconnect', '-l',
        ]

    def output_get_interconnect_id(self):
        return (
            'id name source_server destination_system destination_server\n'
            '%(id)s  loopback  %(src_server)s  nas149   %(dest_server)s\n' %
            {'id': self.interconnect_id,
             'src_server': self.mover_name,
             'dest_server': self.mover_name}
        )

    def cmd_get_physical_devices(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/server_sysconfig',
            self.mover_name, '-pci',
        ]

    def output_get_physical_devices(self):
        return (
            'server_2 : PCI DEVICES:\n'
            'On Board:\n'
            '  PMC QE8 Fibre Channel Controller\n'
            '    0:  fcp-0-0  IRQ: 20 addr: 5006016047a00245\n'
            '    0:  fcp-0-1  IRQ: 21 addr: 5006016147a00245\n'
            '    0:  fcp-0-2  IRQ: 22 addr: 5006016247a00245\n'
            '    0:  fcp-0-3  IRQ: 23 addr: 5006016347a00245\n'
            '  Broadcom Gigabit Ethernet Controller\n'
            '    0:  cge-1-0  IRQ: 24\n'
            '    speed=auto duplex=auto txflowctl=disable rxflowctl=disable\n'
            '    Link: Up\n'
            '    0:  cge-1-1  IRQ: 25\n'
            '    speed=auto duplex=auto txflowctl=disable rxflowctl=disable\n'
            '    Link: Down\n'
            '    0:  cge-1-2  IRQ: 26\n'
            '    speed=auto duplex=auto txflowctl=disable rxflowctl=disable\n'
            '    Link: Down\n'
            '    0:  cge-1-3  IRQ: 27\n'
            '    speed=auto duplex=auto txflowctl=disable rxflowctl=disable\n'
            '    Link: Up\n'
            'Slot: 4\n'
            '  PLX PCI-Express Switch  Controller\n'
            '    1:  PLX PEX8648  IRQ: 10\n'
        )


class DNSDomainTestData(StorageObjectTestData):
    def __init__(self):
        super(DNSDomainTestData, self).__init__()

    @start_task
    def req_create(self, ip_addr=None):
        if ip_addr is None:
            ip_addr = self.dns_ip_address

        return (
            '<NewMoverDnsDomain mover="%(mover_id)s" protocol="udp" '
            'name="%(domain_name)s" servers="%(server_ips)s"/>' %
            {'mover_id': self.mover_id,
             'domain_name': self.domain_name,
             'server_ips': ip_addr}
        )

    @start_task
    def req_delete(self):
        return (
            '<DeleteMoverDnsDomain mover="%(mover_id)s" '
            'name="%(domain_name)s"/>' %
            {'mover_id': self.mover_id,
             'domain_name': self.domain_name}
        )


class CIFSServerTestData(StorageObjectTestData):
    def __init__(self):
        super(CIFSServerTestData, self).__init__()

    @start_task
    def req_create(self, mover_id, is_vdm=True, ip_addr=None):
        if ip_addr is None:
            ip_addr = self.ip_address1

        return (
            '<NewW2KCifsServer interfaces="%(ip)s" compName="%(comp_name)s" '
            'name="%(name)s" domain="%(domain)s">'
            '<MoverOrVdm mover="%(mover_id)s" moverIdIsVdm="%(is_vdm)s"/>'
            '<Aliases><li>%(alias)s</li></Aliases>'
            '<JoinDomain userName="%(domain_user)s" '
            'password="%(domain_password)s"/>'
            '</NewW2KCifsServer>'
            % {'ip': ip_addr,
               'comp_name': self.cifs_server_name,
               'name': self.cifs_server_name[-14:],
               'mover_id': mover_id,
               'alias': self.cifs_server_name[-12:],
               'domain_user': self.domain_user,
               'domain_password': self.domain_password,
               'domain': self.domain_name,
               'is_vdm': 'true' if is_vdm else 'false'}
        )

    @query
    def req_get(self, mover_id, is_vdm=True):
        return (
            '<CifsServerQueryParams>'
            '<MoverOrVdm mover="%(mover_id)s" moverIdIsVdm="%(is_vdm)s"/>'
            '</CifsServerQueryParams>' %
            {'mover_id': mover_id,
             'is_vdm': 'true' if is_vdm else 'false'}
        )

    @response
    def resp_get_succeed(self, mover_id, is_vdm, join_domain,
                         cifs_server_name=None,
                         ip_addr=None):
        if cifs_server_name is None:
            cifs_server_name = self.cifs_server_name

        if ip_addr is None:
            ip_addr = self.ip_address1

        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<CifsServer interfaces="%(ip)s" type="W2K" '
            'localUsers="false" name="%(cifsserver)s" '
            'mover="%(mover_id)s" moverIdIsVdm="%(is_vdm)s"><Aliases>'
            '<li>%(alias)s</li></Aliases><W2KServerData domain='
            '"%(domain)s" compName="%(comp_name)s" '
            'domainJoined="%(join_domain)s"/></CifsServer>'
            % {'mover_id': mover_id,
               'cifsserver': self.cifs_server_name[-14:],
               'ip': ip_addr,
               'is_vdm': 'true' if is_vdm else 'false',
               'alias': self.cifs_server_name[-12:],
               'domain': self.domain_name,
               'join_domain': 'true' if join_domain else 'false',
               'comp_name': cifs_server_name}
        )

    @response
    def resp_get_without_interface(self, mover_id, is_vdm, join_domain):
        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<CifsServer interfaces="" type="W2K" localUsers="false" '
            'name="%(cifsserver)s" mover="%(mover_id)s" '
            'moverIdIsVdm="%(is_vdm)s">'
            '<Aliases><li>%(alias)s</li></Aliases>'
            '<W2KServerData domain="%(domain)s" compName="%(comp_name)s" '
            'domainJoined="%(join_domain)s"/></CifsServer>'
            % {'mover_id': mover_id,
               'cifsserver': self.cifs_server_name[-14:],
               'is_vdm': 'true' if is_vdm else 'false',
               'alias': self.cifs_server_name[-12:],
               'domain': self.domain_name,
               'join_domain': 'true' if join_domain else 'false',
               'comp_name': self.cifs_server_name}
        )

    @start_task
    def req_modify(self, mover_id, is_vdm=True, join_domain=False):
        return (
            '<ModifyW2KCifsServer mover="%(mover_id)s" '
            'moverIdIsVdm="%(is_vdm)s" name="%(cifsserver)s">'
            '<DomainSetting userName="%(username)s" password="%(pw)s" '
            'joinDomain="%(join_domain)s"/>'
            '</ModifyW2KCifsServer>'
            % {'mover_id': mover_id,
               'is_vdm': 'true' if is_vdm else 'false',
               'join_domain': 'true' if join_domain else 'false',
               'cifsserver': self.cifs_server_name[-14:],
               'username': self.domain_user,
               'pw': self.domain_password}
        )

    @response
    def resp_modify_but_already_join_domain(self):
        return (
            '<Status maxSeverity="error"> '
            '<Problem messageCode="13157007726" component="DART" '
            'message="Fake message." severity="error">'
            '<Description>Fake description</Description>'
            '<Action>Fake action.</Action>'
            '</Problem>'
            '</Status> '
        )

    @response
    def resp_modify_but_unjoin_domain(self):
        return (
            '<Status maxSeverity="error"> '
            '<Problem messageCode="13157007723" component="DART" '
            'message="Fake message." severity="error">'
            '<Description>Fake description</Description>'
            '<Action>Fake action.</Action>'
            '</Problem>'
            '</Status> '
        )

    @start_task
    def req_delete(self, mover_id, is_vdm=True):
        return (
            '<DeleteCifsServer mover="%(mover_id)s" moverIdIsVdm="%(is_vdm)s" '
            'name="%(cifsserver)s"/>'
            % {'mover_id': mover_id,
               'is_vdm': 'true' if is_vdm else 'false',
               'cifsserver': self.cifs_server_name[-14:]}
        )


class CIFSShareTestData(StorageObjectTestData):
    def __init__(self):
        super(CIFSShareTestData, self).__init__()

    @start_task
    def req_create(self, mover_id, is_vdm=True):
        return (
            '<NewCifsShare path="%(path)s" name="%(share_name)s">'
            '<MoverOrVdm mover="%(mover_id)s" moverIdIsVdm="%(is_vdm)s"/>'
            '<CifsServers><li>%(cifsserver)s</li></CifsServers>'
            '</NewCifsShare>' %
            {'path': '/' + self.share_name,
             'share_name': self.share_name,
             'mover_id': mover_id,
             'is_vdm': 'true' if is_vdm else 'false',
             'cifsserver': self.cifs_server_name[-14:]}
        )

    @start_task
    def req_delete(self, mover_id, is_vdm=True):
        return (
            '<DeleteCifsShare mover="%(mover_id)s" moverIdIsVdm="%(is_vdm)s" '
            'name="%(share_name)s">'
            '<CifsServers><li>%(cifsserver)s</li></CifsServers>'
            '</DeleteCifsShare>' %
            {'share_name': self.share_name,
             'mover_id': mover_id,
             'is_vdm': 'true' if is_vdm else 'false',
             'cifsserver': self.cifs_server_name[-12:]}
        )

    @query
    def req_get(self):
        return '<CifsShareQueryParams name="%s"/>' % self.share_name

    @response
    def resp_get_succeed(self, mover_id, is_vdm=True):
        return (
            '<QueryStatus maxSeverity="ok"/>'
            '<CifsShare path="%(path)s" fileSystem="%(fsid)s" name="%(name)s" '
            'mover="%(moverid)s" moverIdIsVdm="%(is_vdm)s">'
            '<CifsServers><li>%(alias)s</li>'
            '</CifsServers>'
            '</CifsShare>' %
            {'path': self.path,
             'fsid': self.filesystem_id,
             'name': self.share_name,
             'moverid': mover_id,
             'is_vdm': 'true' if is_vdm else 'false',
             'alias': self.cifs_server_name[-12:]}
        )

    def cmd_disable_access(self):
        cmd_str = 'sharesd %s set noaccess' % self.share_name
        return [
            'env', 'NAS_DB=/nas',
            '/nas/bin/.server_config', self.vdm_name,
            '-v', '%s' % cmd_str,
        ]

    def cmd_change_access(self, access_level=const.ACCESS_LEVEL_RW,
                          action='grant', user=None):
        if user is None:
            user = self.domain_user
        account = user + '@' + self.domain_name

        if access_level == const.ACCESS_LEVEL_RW:
            str_access = 'fullcontrol'
        else:
            str_access = 'read'

        allow_str = (
            'sharesd %(share_name)s %(action)s %(account)s=%(access)s'
            % {'share_name': self.share_name,
               'action': action,
               'account': account,
               'access': str_access}
        )
        return [
            'env', 'NAS_DB=/nas',
            '/nas/bin/.server_config', self.vdm_name,
            '-v', '%s' % allow_str,
        ]

    def cmd_get_access(self):
        get_str = 'sharesd %s dump' % self.share_name
        return [
            'env', 'NAS_DB=/nas',
            '/nas/bin/.server_config', self.vdm_name,
            '-v', '%s' % get_str,
        ]

    def output_allow_access(self):
        return (
            "Command succeeded:  :3 sharesd %(share)s grant             "
            "%(user)s@%(domain)s=fullcontrol"
            % {'share': self.share_name,
               'user': self.domain_user,
               'domain': self.domain_name}
        )

    def output_allow_access_but_duplicate_ace(self):
        return (
            '%(vdm_name)s : commands processed: 1'
            'output is complete'
            '1443422844: SMB: 6: ACE for %(domain)s\\%(user)s '
            'unchanged'
            '1443422844: ADMIN: 3: '
            'Command failed:  :23 '
            'sharesd %(share)s grant %(user)s@%(domain)s=read'
            'Error 4020: %(vdm_name)s : failed to complete command"'
            % {'share': self.share_name,
               'user': self.domain_user,
               'domain': self.domain_name,
               'vdm_name': self.vdm_name}
        )

    def output_deny_access_but_no_ace(self):
        return (
            '%(vdm_name)s : commands processed: 1'
            'output is complete'
            '1443515516: SMB: 6: No ACE found for %(domain)s\\%(user)s '
            '1443515516: ADMIN: 3: '
            'Command failed:  :26 '
            'sharesd %(share)s revoke %(user)s@%(domain)s=read'
            'Error 4020: %(vdm_name)s : failed to complete command"'
            % {'share': self.share_name,
               'user': self.domain_user,
               'domain': self.domain_name,
               'vdm_name': self.vdm_name}
        )

    def output_deny_access_but_no_user_found(self):
        return (
            '%(vdm_name)s : commands processed: 1'
            'output is complete'
            '1443520322: SMB: 6: Cannot get mapping for %(domain)s\\%(user)s '
            '1443520322: ADMIN: 3: '
            'Command failed:  :26 '
            'sharesd %(share)s revoke %(user)s@%(domain)s=read'
            'Error 4020: %(vdm_name)s : failed to complete command"'
            % {'share': self.share_name,
               'user': self.domain_user,
               'domain': self.domain_name,
               'vdm_name': self.vdm_name}
        )


class NFSShareTestData(StorageObjectTestData):
    def __init__(self):
        super(NFSShareTestData, self).__init__()

    def cmd_create(self):
        default_access = 'access=-0.0.0.0/0.0.0.0'
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', self.vdm_name,
            '-option', default_access,
            self.path,
        ]

    def output_create(self):
        return "%s : done" % self.vdm_name

    def cmd_get(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', self.vdm_name,
            '-P', 'nfs',
            '-list', self.path,
        ]

    def output_get_succeed(self, rw_hosts, ro_hosts):
        rw_hosts = [utils.convert_ipv6_format_if_needed(ip_addr) for ip_addr in
                    rw_hosts]
        ro_hosts = [utils.convert_ipv6_format_if_needed(ip_addr) for ip_addr in
                    ro_hosts]

        if rw_hosts and ro_hosts:
            return (
                '%(mover_name)s :\nexport "%(path)s" '
                'access=-0.0.0.0/0.0.0.0:%(host)s root=%(host)s '
                'rw=%(rw_host)s ro=%(ro_host)s\n'
                % {'mover_name': self.vdm_name,
                   'path': self.path,
                   'host': ":".join(rw_hosts + ro_hosts),
                   'rw_host': ":".join(rw_hosts),
                   'ro_host': ":".join(ro_hosts)}
            )
        elif rw_hosts:
            return (
                '%(mover_name)s :\nexport "%(path)s" '
                'access=-0.0.0.0/0.0.0.0:%(host)s root=%(host)s '
                'rw=%(rw_host)s\n'
                % {'mover_name': self.vdm_name,
                   'host': ":".join(rw_hosts),
                   'path': self.path,
                   'rw_host': ":".join(rw_hosts)}
            )
        elif ro_hosts:
            return (
                '%(mover_name)s :\nexport "%(path)s" '
                'access=-0.0.0.0/0.0.0.0:%(host)s root=%(host)s '
                'ro=%(ro_host)s\n'
                % {'mover_name': self.vdm_name,
                   'host': ":".join(ro_hosts),
                   'path': self.path,
                   'ro_host': ":".join(ro_hosts)}
            )
        else:
            return (
                '%(mover_name)s :\nexport "%(path)s" '
                'access=-0.0.0.0/0.0.0.0\n'
                % {'mover_name': self.vdm_name,
                   'path': self.path}
            )

    def output_get_but_not_found(self):
        return (
            '%(mover_name)s : \nError 2: %(mover_name)s : '
            'No such file or directory \n' % {'mover_name': self.vdm_name}
        )

    def cmd_delete(self):
        return [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', self.vdm_name,
            '-unexport',
            '-perm',
            self.path,
        ]

    def output_delete_succeed(self):
        return "%s : done" % self.vdm_name

    def output_delete_but_locked(self):
        return ("Error 2201: %s : unable to acquire lock(s), try later"
                % self.vdm_name)

    def cmd_set_access(self, rw_hosts, ro_hosts):
        rw_hosts = [utils.convert_ipv6_format_if_needed(ip_addr) for ip_addr in
                    rw_hosts]
        ro_hosts = [utils.convert_ipv6_format_if_needed(ip_addr) for ip_addr in
                    ro_hosts]

        access_str = ("access=-0.0.0.0/0.0.0.0:%(access_hosts)s,"
                      "root=%(root_hosts)s,rw=%(rw_hosts)s,ro=%(ro_hosts)s" %
                      {'rw_hosts': ":".join(rw_hosts),
                       'ro_hosts': ":".join(ro_hosts),
                       'root_hosts': ":".join(rw_hosts + ro_hosts),
                       'access_hosts': ":".join(rw_hosts + ro_hosts)})

        return [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', self.vdm_name,
            '-ignore',
            '-option', access_str,
            self.path,
        ]

    def output_set_access_success(self):
        return "%s : done" % self.vdm_name


class FakeEMCShareDriver(object):
    def __init__(self, enas_type='vnx'):
        self.configuration = conf.Configuration(None)
        self.configuration.append_config_values = mock.Mock(return_value=0)
        self.configuration.emc_share_backend = FakeData.emc_share_backend
        self.configuration.vnx_server_container = FakeData.mover_name
        if enas_type == 'vmax':
            self.configuration.emc_share_backend = FakeData.vmax_share_backend
            self.configuration.vmax_server_container = FakeData.mover_name
        self.configuration.emc_nas_server = FakeData.emc_nas_server
        self.configuration.emc_nas_login = FakeData.emc_nas_login
        self.configuration.emc_nas_password = FakeData.emc_nas_password
        self.configuration.share_backend_name = FakeData.share_backend_name

CIFS_SHARE = fake_share.fake_share(
    id=FakeData.share_id,
    name=FakeData.share_name,
    size=FakeData.share_size,
    share_network_id=FakeData.share_network_id,
    share_server_id=FakeData.share_server_id,
    host=FakeData.host,
    share_proto='CIFS')

NFS_SHARE = fake_share.fake_share(
    id=FakeData.share_id,
    name=FakeData.share_name,
    size=FakeData.share_size,
    share_network_id=FakeData.share_network_id,
    share_server_id=FakeData.share_server_id,
    host=FakeData.host,
    share_proto='NFS')

CIFS_RW_ACCESS = fake_share.fake_access(
    access_type='user',
    access_to=FakeData.domain_user,
    access_level='rw')

CIFS_RO_ACCESS = fake_share.fake_access(
    access_type='user',
    access_to=FakeData.domain_user,
    access_level='ro')

NFS_RW_ACCESS = fake_share.fake_access(
    access_type='ip',
    access_to=FakeData.nfs_host_ip,
    access_level='rw')

NFS_RW_ACCESS_IPV6 = fake_share.fake_access(
    access_type='ip',
    access_to=FakeData.nfs_host_ipv6,
    access_level='rw')

NFS_RO_ACCESS = fake_share.fake_access(
    access_type='ip',
    access_to=FakeData.nfs_host_ip,
    access_level='ro')

NFS_RO_ACCESS_IPV6 = fake_share.fake_access(
    access_type='ip',
    access_to=FakeData.nfs_host_ipv6,
    access_level='ro')

SHARE_SERVER = {
    'id': FakeData.share_server_id,
    'share_network': {
        'name': 'fake_share_network',
        'id': FakeData.share_network_id
    },
    'share_network_id': FakeData.share_network_id,
    'backend_details': {
        'share_server_name': FakeData.vdm_name,
        'cifs_if': FakeData.network_allocations_ip1,
        'nfs_if': FakeData.network_allocations_ip2,
    }
}

SHARE_SERVER_IPV6 = {
    'id': FakeData.share_server_id,
    'share_network': {
        'name': 'fake_share_network',
        'id': FakeData.share_network_id
    },
    'share_network_id': FakeData.share_network_id,
    'backend_details': {
        'share_server_name': FakeData.vdm_name,
        'cifs_if': FakeData.network_allocations_ip3,
        'nfs_if': FakeData.network_allocations_ip4,
    }
}

SERVER_DETAIL = {
    'share_server_name': FakeData.vdm_name,
    'cifs_if': FakeData.network_allocations_ip1,
    'nfs_if': FakeData.network_allocations_ip2,
}

SERVER_DETAIL_IPV6 = {
    'share_server_name': FakeData.vdm_name,
    'cifs_if': FakeData.network_allocations_ip3,
    'nfs_if': FakeData.network_allocations_ip4,
}

SECURITY_SERVICE = [
    {
        'type': 'active_directory',
        'domain': FakeData.domain_name,
        'dns_ip': FakeData.dns_ip_address,
        'user': FakeData.domain_user,
        'password': FakeData.domain_password
    },
]

SECURITY_SERVICE_IPV6 = [
    {
        'type': 'active_directory',
        'domain': FakeData.domain_name,
        'dns_ip': FakeData.dns_ipv6_address,
        'user': FakeData.domain_user,
        'password': FakeData.domain_password
    },
]

NETWORK_INFO = {
    'server_id': FakeData.share_server_id,
    'cidr': FakeData.cidr,
    'security_services': [
        {'type': 'active_directory',
         'domain': FakeData.domain_name,
         'dns_ip': FakeData.dns_ip_address,
         'user': FakeData.domain_user,
         'password': FakeData.domain_password},
    ],
    'segmentation_id': FakeData.segmentation_id,
    'network_type': 'vlan',
    'network_allocations': [
        {'id': FakeData.network_allocations_id1,
         'ip_address': FakeData.network_allocations_ip1,
         'ip_version': FakeData.network_allocations_ip_version1},
        {'id': FakeData.network_allocations_id2,
         'ip_address': FakeData.network_allocations_ip2,
         'ip_version': FakeData.network_allocations_ip_version2}
    ]
}

NETWORK_INFO_IPV6 = {
    'server_id': FakeData.share_server_id,
    'cidr': FakeData.cidr_v6,
    'security_services': [
        {'type': 'active_directory',
         'domain': FakeData.domain_name,
         'dns_ip': FakeData.dns_ipv6_address,
         'user': FakeData.domain_user,
         'password': FakeData.domain_password},
    ],
    'segmentation_id': FakeData.segmentation_id,
    'network_type': 'vlan',
    'network_allocations': [
        {'id': FakeData.network_allocations_id3,
         'ip_address': FakeData.network_allocations_ip3,
         'ip_version': FakeData.network_allocations_ip_version3},
        {'id': FakeData.network_allocations_id4,
         'ip_address': FakeData.network_allocations_ip4,
         'ip_version': FakeData.network_allocations_ip_version4}
    ]
}

STATS = dict(
    share_backend_name='VNX',
    vendor_name='EMC',
    storage_protocol='NFS_CIFS',
    driver_version='2.0.0,')

STATS_VMAX = dict(
    share_backend_name='VMAX',
    vendor_name='EMC',
    storage_protocol='NFS_CIFS',
    driver_version='2.0.0,')



# Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
ZFS Storage Appliance REST API Client Programmatic Interface
TODO(diemtran): this module needs to be placed in a library common to OpenStack
    services. When this happens, the file should be removed from Manila code
    base and imported from the relevant library.
"""

import time

from oslo_serialization import jsonutils
import six
from six.moves import http_client
from six.moves.urllib import error as urlerror
from six.moves.urllib import request as urlrequest


def log_debug_msg(obj, message):
    if obj.log_function:
        obj.log_function(message)


class Status(object):
    """Result HTTP Status."""

    #: Request return OK
    OK = http_client.OK  # pylint: disable=invalid-name

    #: New resource created successfully
    CREATED = http_client.CREATED

    #: Command accepted
    ACCEPTED = http_client.ACCEPTED

    #: Command returned OK but no data will be returned
    NO_CONTENT = http_client.NO_CONTENT

    #: Bad Request
    BAD_REQUEST = http_client.BAD_REQUEST

    #: User is not authorized
    UNAUTHORIZED = http_client.UNAUTHORIZED

    #: The request is not allowed
    FORBIDDEN = http_client.FORBIDDEN

    #: The requested resource was not found
    NOT_FOUND = http_client.NOT_FOUND

    #: The request is not allowed
    NOT_ALLOWED = http_client.METHOD_NOT_ALLOWED

    #: Request timed out
    TIMEOUT = http_client.REQUEST_TIMEOUT

    #: Invalid request
    CONFLICT = http_client.CONFLICT

    #: Service Unavailable
    BUSY = http_client.SERVICE_UNAVAILABLE


class RestResult(object):
    """Result from a REST API operation."""
    def __init__(self, logfunc=None, response=None, err=None):
        """Initialize a RestResult containing the results from a REST call.

        :param logfunc: debug log function.
        :param response: HTTP response.
        :param err: HTTP error.
        """
        self.response = response
        self.log_function = logfunc
        self.error = err
        self.data = ""
        self.status = 0
        if self.response:
            self.status = self.response.getcode()
            result = self.response.read()
            while result:
                self.data += result
                result = self.response.read()

        if self.error:
            self.status = self.error.code
            self.data = http_client.responses[self.status]

        log_debug_msg(self, 'Response code: %s' % self.status)
        log_debug_msg(self, 'Response data: %s' % self.data)

    def get_header(self, name):
        """Get an HTTP header with the given name from the results.

        :param name: HTTP header name.
        :return: The header value or None if no value is found.
        """
        if self.response is None:
            return None
        info = self.response.info()
        return info.getheader(name)


class RestClientError(Exception):
    """Exception for ZFS REST API client errors."""
    def __init__(self, status, name="ERR_INTERNAL", message=None):

        """Create a REST Response exception.

        :param status: HTTP response status.
        :param name: The name of the REST API error type.
        :param message: Descriptive error message returned from REST call.
        """
        super(RestClientError, self).__init__(message)
        self.code = status
        self.name = name
        self.msg = message
        if status in http_client.responses:
            self.msg = http_client.responses[status]

    def __str__(self):
        return "%d %s %s" % (self.code, self.name, self.msg)


class RestClientURL(object):  # pylint: disable=R0902
    """ZFSSA urllib client."""
    def __init__(self, url, logfunc=None, **kwargs):
        """Initialize a REST client.

        :param url: The ZFSSA REST API URL.
        :key session: HTTP Cookie value of x-auth-session obtained from a
                      normal BUI login.
        :key timeout: Time in seconds to wait for command to complete.
                      (Default is 60 seconds).
        """
        self.url = url
        self.log_function = logfunc
        self.local = kwargs.get("local", False)
        self.base_path = kwargs.get("base_path", "/api")
        self.timeout = kwargs.get("timeout", 60)
        self.headers = None
        if kwargs.get('session'):
            self.headers['x-auth-session'] = kwargs.get('session')

        self.headers = {"content-type": "application/json"}
        self.do_logout = False
        self.auth_str = None

    def _path(self, path, base_path=None):
        """Build rest url path."""
        if path.startswith("http://") or path.startswith("https://"):
            return path
        if base_path is None:
            base_path = self.base_path
        if not path.startswith(base_path) and not (
                self.local and ("/api" + path).startswith(base_path)):
            path = "%s%s" % (base_path, path)
        if self.local and path.startswith("/api"):
            path = path[4:]
        return self.url + path

    def _authorize(self):
        """Performs authorization setting x-auth-session."""
        self.headers['authorization'] = 'Basic %s' % self.auth_str
        if 'x-auth-session' in self.headers:
            del self.headers['x-auth-session']

        try:
            result = self.post("/access/v1")
            del self.headers['authorization']
            if result.status == http_client.CREATED:
                self.headers['x-auth-session'] = (
                    result.get_header('x-auth-session'))
                self.do_logout = True
                log_debug_msg(self, ('ZFSSA version: %s')
                              % result.get_header('x-zfssa-version'))

            elif result.status == http_client.NOT_FOUND:
                raise RestClientError(result.status, name="ERR_RESTError",
                                      message=("REST Not Available:"
                                               "Please Upgrade"))

        except RestClientError:
            del self.headers['authorization']
            raise

    def login(self, auth_str):
        """Login to an appliance using a user name and password.

        Start a session like what is done logging into the BUI.  This is not a
        requirement to run REST commands, since the protocol is stateless.
        What is does is set up a cookie session so that some server side
        caching can be done.  If login is used remember to call logout when
        finished.

        :param auth_str: Authorization string (base64).
        """
        self.auth_str = auth_str
        self._authorize()

    def logout(self):
        """Logout of an appliance."""
        result = None
        try:
            result = self.delete("/access/v1", base_path="/api")
        except RestClientError:
            pass

        self.headers.clear()
        self.do_logout = False
        return result

    def islogin(self):
        """return if client is login."""
        return self.do_logout

    @staticmethod
    def mkpath(*args, **kwargs):
        """Make a path?query string for making a REST request.

        :cmd_params args: The path part.
        :cmd_params kwargs: The query part.
        """
        buf = six.StringIO()
        query = "?"
        for arg in args:
            buf.write("/")
            buf.write(arg)
        for k in kwargs:
            buf.write(query)
            if query == "?":
                query = "&"
            buf.write(k)
            buf.write("=")
            buf.write(kwargs[k])
        return buf.getvalue()

    # pylint: disable=R0912
    def request(self, path, request, body=None, **kwargs):
        """Make an HTTP request and return the results.

        :param path: Path used with the initialized URL to make a request.
        :param request: HTTP request type (GET, POST, PUT, DELETE).
        :param body: HTTP body of request.
        :key accept: Set HTTP 'Accept' header with this value.
        :key base_path: Override the base_path for this request.
        :key content: Set HTTP 'Content-Type' header with this value.
        """
        out_hdrs = dict.copy(self.headers)
        if kwargs.get("accept"):
            out_hdrs['accept'] = kwargs.get("accept")

        if body:
            if isinstance(body, dict):
                body = six.text_type(jsonutils.dumps(body))

        if body and len(body):
            out_hdrs['content-length'] = len(body)

        zfssaurl = self._path(path, kwargs.get("base_path"))
        req = urlrequest.Request(zfssaurl, body, out_hdrs)
        req.get_method = lambda: request
        maxreqretries = kwargs.get("maxreqretries", 10)
        retry = 0
        response = None

        log_debug_msg(self, 'Request: %s %s' % (request, zfssaurl))
        log_debug_msg(self, 'Out headers: %s' % out_hdrs)
        if body and body != '':
            log_debug_msg(self, 'Body: %s' % body)

        while retry < maxreqretries:
            try:
                response = urlrequest.urlopen(req, timeout=self.timeout)
            except urlerror.HTTPError as err:
                if err.code == http_client.NOT_FOUND:
                    log_debug_msg(self, 'REST Not Found: %s' % err.code)
                else:
                    log_debug_msg(self, ('REST Not Available: %s') % err.code)

                if (err.code == http_client.SERVICE_UNAVAILABLE and
                        retry < maxreqretries):
                    retry += 1
                    time.sleep(1)
                    log_debug_msg(self, ('Server Busy retry request: %s')
                                  % retry)
                    continue
                if ((err.code == http_client.UNAUTHORIZED or
                     err.code == http_client.INTERNAL_SERVER_ERROR) and
                        '/access/v1' not in zfssaurl):
                    try:
                        log_debug_msg(self, ('Authorizing request: '
                                             '%(zfssaurl)s '
                                             'retry: %(retry)d .')
                                      % {'zfssaurl': zfssaurl,
                                         'retry': retry})
                        self._authorize()
                        req.add_header('x-auth-session',
                                       self.headers['x-auth-session'])
                    except RestClientError:
                        log_debug_msg(self, ('Cannot authorize.'))
                    retry += 1
                    time.sleep(1)
                    continue

                return RestResult(self.log_function, err=err)

            except urlerror.URLError as err:
                log_debug_msg(self, ('URLError: %s') % err.reason)
                raise RestClientError(-1, name="ERR_URLError",
                                      message=err.reason)
            break

        if ((response and
             response.getcode() == http_client.SERVICE_UNAVAILABLE) and
                retry >= maxreqretries):
            raise RestClientError(response.getcode(), name="ERR_HTTPError",
                                  message="REST Not Available: Disabled")

        return RestResult(self.log_function, response=response)

    def get(self, path, **kwargs):
        """Make an HTTP GET request.

        :param path: Path to resource.
        """
        return self.request(path, "GET", **kwargs)

    def post(self, path, body="", **kwargs):
        """Make an HTTP POST request.

        :param path: Path to resource.
        :param body: Post data content.
        """
        return self.request(path, "POST", body, **kwargs)

    def put(self, path, body="", **kwargs):
        """Make an HTTP PUT request.

        :param path: Path to resource.
        :param body: Put data content.
        """
        return self.request(path, "PUT", body, **kwargs)

    def delete(self, path, **kwargs):
        """Make an HTTP DELETE request.

        :param path: Path to resource that will be deleted.
        """
        return self.request(path, "DELETE", **kwargs)

    def head(self, path, **kwargs):
        """Make an HTTP HEAD request.

        :param path: Path to resource.
        """
        return self.request(path, "HEAD", **kwargs)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from manila import context
from manila.db import api as db_api
from manila.scheduler.weighers import base_host
from manila.share import utils

pool_weight_opts = [
    cfg.FloatOpt('pool_weight_multiplier',
                 default=1.0,
                 help='Multiplier used for weighing pools which have '
                      'existing share servers. Negative numbers mean to spread'
                      ' vs stack.'),
]

CONF = cfg.CONF
CONF.register_opts(pool_weight_opts)


class PoolWeigher(base_host.BaseHostWeigher):
    def weight_multiplier(self):
        """Override the weight multiplier."""
        return CONF.pool_weight_multiplier

    def _weigh_object(self, host_state, weight_properties):
        """Pools with existing share server win."""
        pool_mapping = weight_properties.get('server_pools_mapping', {})
        if not pool_mapping:
            return 0

        ctx = context.get_admin_context()
        host = utils.extract_host(host_state.host, 'backend')
        servers = db_api.share_server_get_all_by_host(ctx, host)
        pool = utils.extract_host(host_state.host, 'pool')
        for server in servers:
            if any(pool == p['pool_name'] for p in pool_mapping.get(
                   server['id'], [])):
                return 1
        return 0



# Copyright 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
import webob

from manila.api import common
from manila.api.v1 import share_manage
from manila.db import api as db_api
from manila import exception
from manila import policy
from manila.share import api as share_api
from manila.share import share_types
from manila import test
from manila.tests.api import fakes
from manila import utils


def get_fake_manage_body(export_path='/fake', service_host='fake@host#POOL',
                         protocol='fake', share_type='fake', **kwargs):
    fake_share = {
        'export_path': export_path,
        'service_host': service_host,
        'protocol': protocol,
        'share_type': share_type,
    }
    fake_share.update(kwargs)
    return {'share': fake_share}


@ddt.ddt
class ShareManageTest(test.TestCase):
    """Share Manage Test."""
    def setUp(self):
        super(ShareManageTest, self).setUp()
        self.controller = share_manage.ShareManageController()
        self.resource_name = self.controller.resource_name
        self.request = fakes.HTTPRequest.blank('/share/manage',
                                               use_admin_context=True)
        self.context = self.request.environ['manila.context']
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.mock_object(
            common, 'validate_public_share_policy',
            mock.Mock(side_effect=lambda *args, **kwargs: args[1]))

    @ddt.data({},
              {'shares': {}},
              {'share': get_fake_manage_body('', None, None)})
    def test_share_manage_invalid_body(self, body):
        self.assertRaises(webob.exc.HTTPUnprocessableEntity,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_service_not_found(self):
        body = get_fake_manage_body()
        self.mock_object(db_api, 'service_get_by_host_and_topic', mock.Mock(
            side_effect=exception.ServiceNotFound(service_id='fake')))

        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_share_type_not_found(self):
        body = get_fake_manage_body()
        self.mock_object(db_api, 'service_get_by_host_and_topic', mock.Mock())
        self.mock_object(utils, 'service_is_up', mock.Mock(return_value=True))
        self.mock_object(db_api, 'share_type_get_by_name', mock.Mock(
            side_effect=exception.ShareTypeNotFoundByName(
                share_type_name='fake')))

        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def _setup_manage_mocks(self, service_is_up=True):
        self.mock_object(db_api, 'service_get_by_host_and_topic', mock.Mock(
            return_value={'host': 'fake'}))
        self.mock_object(share_types, 'get_share_type_by_name_or_id',
                         mock.Mock(return_value={'id': 'fake'}))
        self.mock_object(utils, 'service_is_up', mock.Mock(
            return_value=service_is_up))

    @ddt.data({'service_is_up': False, 'service_host': 'fake@host#POOL'},
              {'service_is_up': True, 'service_host': 'fake@host'})
    def test_share_manage_bad_request(self, settings):
        body = get_fake_manage_body(service_host=settings.pop('service_host'))
        self._setup_manage_mocks(**settings)

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_duplicate_share(self):
        body = get_fake_manage_body()
        exc = exception.InvalidShare(reason="fake")
        self._setup_manage_mocks()
        self.mock_object(share_api.API, 'manage', mock.Mock(side_effect=exc))

        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_forbidden_manage(self):
        body = get_fake_manage_body()
        self._setup_manage_mocks()
        error = mock.Mock(side_effect=exception.PolicyNotAuthorized(action=''))
        self.mock_object(share_api.API, 'manage', error)

        self.assertRaises(webob.exc.HTTPForbidden,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_forbidden_validate_service_host(self):
        body = get_fake_manage_body()
        self._setup_manage_mocks()
        error = mock.Mock(side_effect=exception.PolicyNotAuthorized(action=''))
        self.mock_object(utils, 'service_is_up', mock.Mock(side_effect=error))

        self.assertRaises(webob.exc.HTTPForbidden,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_invalid_input(self):
        body = get_fake_manage_body()
        self._setup_manage_mocks()
        error = mock.Mock(side_effect=exception.InvalidInput(message="",
                                                             reason="fake"))
        self.mock_object(share_api.API, 'manage', mock.Mock(side_effect=error))

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_invalid_share_server(self):
        body = get_fake_manage_body()
        self._setup_manage_mocks()
        error = mock.Mock(
            side_effect=exception.InvalidShareServer(message="",
                                                     share_server_id="")
        )
        self.mock_object(share_api.API, 'manage', mock.Mock(side_effect=error))

        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller.create,
                          self.request,
                          body)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    @ddt.data(
        get_fake_manage_body(name='foo', description='bar'),
        get_fake_manage_body(display_name='foo', description='bar'),
        get_fake_manage_body(name='foo', display_description='bar'),
        get_fake_manage_body(display_name='foo', display_description='bar'),
        get_fake_manage_body(display_name='foo', display_description='bar',
                             driver_options=dict(volume_id='quuz')),
    )
    def test_share_manage(self, data):
        self._setup_manage_mocks()
        return_share = {'share_type_id': '', 'id': 'fake'}
        self.mock_object(
            share_api.API, 'manage', mock.Mock(return_value=return_share))
        share = {
            'host': data['share']['service_host'],
            'export_location': data['share']['export_path'],
            'share_proto': data['share']['protocol'].upper(),
            'share_type_id': 'fake',
            'display_name': 'foo',
            'display_description': 'bar',
        }
        data['share']['is_public'] = 'foo'
        driver_options = data['share'].get('driver_options', {})

        actual_result = self.controller.create(self.request, data)

        share_api.API.manage.assert_called_once_with(
            mock.ANY, share, driver_options)
        self.assertIsNotNone(actual_result)
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_share_manage_allow_dhss_true(self):
        self._setup_manage_mocks()
        data = get_fake_manage_body(name='foo', description='bar')
        return_share = {'share_type_id': '', 'id': 'fake'}
        self.mock_object(
            share_api.API, 'manage', mock.Mock(return_value=return_share))
        share = {
            'host': data['share']['service_host'],
            'export_location': data['share']['export_path'],
            'share_proto': data['share']['protocol'].upper(),
            'share_type_id': 'fake',
            'display_name': 'foo',
            'display_description': 'bar',
            'share_server_id': 'fake'
        }
        data['share']['share_server_id'] = 'fake'
        driver_options = data['share'].get('driver_options', {})

        self.controller._manage(self.request, data, allow_dhss_true=True)

        share_api.API.manage.assert_called_once_with(
            self.context, share, driver_options
        )
        self.mock_policy_check.assert_called_once_with(
            self.context, self.resource_name, 'manage')

    def test_wrong_permissions(self):
        body = get_fake_manage_body()
        fake_req = fakes.HTTPRequest.blank(
            '/share/manage', use_admin_context=False)

        self.assertRaises(webob.exc.HTTPForbidden,
                          self.controller.create,
                          fake_req, body)
        self.mock_policy_check.assert_called_once_with(
            fake_req.environ['manila.context'], self.resource_name, 'manage')



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Scheduler Manager
"""

try:
    # Python3 variant
    from importlib import reload
except ImportError:
    pass

import ddt
import mock
from oslo_config import cfg

from manila.common import constants
from manila import context
from manila import db
from manila import exception
from manila.message import message_field
from manila import quota
from manila.scheduler.drivers import base
from manila.scheduler.drivers import filter
from manila.scheduler import manager
from manila.share import rpcapi as share_rpcapi
from manila import test
from manila.tests import db_utils
from manila.tests import fake_share as fakes

CONF = cfg.CONF


@ddt.ddt
class SchedulerManagerTestCase(test.TestCase):
    """Test case for scheduler manager."""

    driver_cls = base.Scheduler
    driver_cls_name = 'manila.scheduler.drivers.base.Scheduler'

    def setUp(self):
        super(SchedulerManagerTestCase, self).setUp()
        self.periodic_tasks = []

        def _periodic_task(*args, **kwargs):
            def decorator(f):
                self.periodic_tasks.append(f)
                return f
            return mock.Mock(side_effect=decorator)

        self.mock_periodic_task = self.mock_object(
            manager.periodic_task, 'periodic_task',
            mock.Mock(side_effect=_periodic_task))
        reload(manager)

        self.flags(scheduler_driver=self.driver_cls_name)
        self.manager = manager.SchedulerManager()
        self.context = context.RequestContext('fake_user', 'fake_project')
        self.topic = 'fake_topic'
        self.fake_args = (1, 2, 3)
        self.fake_kwargs = {'cat': 'meow', 'dog': 'woof'}

    def raise_no_valid_host(self, *args, **kwargs):
            raise exception.NoValidHost(reason="")

    def test_1_correct_init(self):
        # Correct scheduler driver
        manager = self.manager
        self.assertIsInstance(manager.driver, self.driver_cls)

    @ddt.data('manila.scheduler.filter_scheduler.FilterScheduler',
              'manila.scheduler.drivers.filter.FilterScheduler')
    def test_scheduler_driver_mapper(self, driver_class):

        test_manager = manager.SchedulerManager(scheduler_driver=driver_class)

        self.assertIsInstance(test_manager.driver, filter.FilterScheduler)

    def test_init_host(self):

        self.mock_object(context,
                         'get_admin_context',
                         mock.Mock(return_value='fake_admin_context'))
        self.mock_object(self.manager, 'request_service_capabilities')

        self.manager.init_host()

        self.manager.request_service_capabilities.assert_called_once_with(
            'fake_admin_context')

    def test_get_host_list(self):

        self.mock_object(self.manager.driver, 'get_host_list')

        self.manager.get_host_list(context)

        self.manager.driver.get_host_list.assert_called_once_with()

    def test_get_service_capabilities(self):

        self.mock_object(self.manager.driver, 'get_service_capabilities')

        self.manager.get_service_capabilities(context)

        self.manager.driver.get_service_capabilities.assert_called_once_with()

    def test_update_service_capabilities(self):
        service_name = 'fake_service'
        host = 'fake_host'
        with mock.patch.object(self.manager.driver,
                               'update_service_capabilities', mock.Mock()):
            self.manager.update_service_capabilities(
                self.context, service_name=service_name, host=host)
            (self.manager.driver.update_service_capabilities.
                assert_called_once_with(service_name, host, {}))
        with mock.patch.object(self.manager.driver,
                               'update_service_capabilities', mock.Mock()):
            capabilities = {'fake_capability': 'fake_value'}
            self.manager.update_service_capabilities(
                self.context, service_name=service_name, host=host,
                capabilities=capabilities)
            (self.manager.driver.update_service_capabilities.
                assert_called_once_with(service_name, host, capabilities))

    @mock.patch.object(db, 'share_update', mock.Mock())
    @mock.patch('manila.message.api.API.create')
    def test_create_share_exception_puts_share_in_error_state(
            self, _mock_message_create):
        """Test NoValidHost exception for create_share.

        Puts the share in 'error' state and eats the exception.
        """
        fake_share_id = 1

        request_spec = {'share_id': fake_share_id}
        ex = exception.NoValidHost(reason='')
        with mock.patch.object(
                self.manager.driver, 'schedule_create_share',
                mock.Mock(side_effect=ex)):
            self.mock_object(manager.LOG, 'error')

            self.manager.create_share_instance(
                self.context, request_spec=request_spec, filter_properties={})

            db.share_update.assert_called_once_with(
                self.context, fake_share_id, {'status': 'error'})
            (self.manager.driver.schedule_create_share.
                assert_called_once_with(self.context, request_spec, {}))
            manager.LOG.error.assert_called_once_with(mock.ANY, mock.ANY)

            _mock_message_create.assert_called_once_with(
                self.context,
                message_field.Action.ALLOCATE_HOST,
                self.context.project_id, resource_type='SHARE',
                exception=ex, resource_id=fake_share_id)

    @mock.patch.object(db, 'share_update', mock.Mock())
    def test_create_share_other_exception_puts_share_in_error_state(self):
        """Test any exception except NoValidHost for create_share.

        Puts the share in 'error' state and re-raises the exception.
        """
        fake_share_id = 1

        request_spec = {'share_id': fake_share_id}
        with mock.patch.object(self.manager.driver,
                               'schedule_create_share',
                               mock.Mock(side_effect=exception.QuotaError)):
            self.mock_object(manager.LOG, 'error')

            self.assertRaises(exception.QuotaError,
                              self.manager.create_share_instance,
                              self.context,
                              request_spec=request_spec,
                              filter_properties={})

            db.share_update.assert_called_once_with(
                self.context, fake_share_id, {'status': 'error'})
            (self.manager.driver.schedule_create_share.
                assert_called_once_with(self.context, request_spec, {}))
            manager.LOG.error.assert_called_once_with(mock.ANY, mock.ANY)

    @mock.patch.object(quota.QUOTAS, 'expire')
    def test__expire_reservations(self, mock_expire):
        self.manager._expire_reservations(self.context)

        mock_expire.assert_called_once_with(self.context)

    @mock.patch('manila.message.api.API.cleanup_expired_messages')
    def test__clean_expired_messages(self, mock_expire):
        self.manager._clean_expired_messages(self.context)

        mock_expire.assert_called_once_with(self.context)

    def test_periodic_tasks(self):
        self.assertEqual(2, self.mock_periodic_task.call_count)

        self.assertEqual(2, len(self.periodic_tasks))
        self.assertEqual(
            self.periodic_tasks[0].__name__,
            self.manager._expire_reservations.__name__)
        self.assertEqual(
            self.periodic_tasks[1].__name__,
            self.manager._clean_expired_messages.__name__)

    def test_get_pools(self):
        """Ensure get_pools exists and calls base_scheduler.get_pools."""
        mock_get_pools = self.mock_object(self.manager.driver,
                                          'get_pools',
                                          mock.Mock(return_value='fake_pools'))

        result = self.manager.get_pools(self.context, filters='fake_filters')

        mock_get_pools.assert_called_once_with(self.context, 'fake_filters',
                                               False)
        self.assertEqual('fake_pools', result)

    @mock.patch.object(db, 'share_group_update', mock.Mock())
    def test_create_group_no_valid_host_puts_group_in_error_state(self):
        """Test that NoValidHost is raised for create_share_group.

        Puts the share in 'error' state and eats the exception.
        """

        fake_group_id = 1
        group_id = fake_group_id
        request_spec = {"share_group_id": group_id}
        with mock.patch.object(
                self.manager.driver, 'schedule_create_share_group',
                mock.Mock(side_effect=self.raise_no_valid_host)):
            self.manager.create_share_group(self.context,
                                            fake_group_id,
                                            request_spec=request_spec,
                                            filter_properties={})
            db.share_group_update.assert_called_once_with(
                self.context, fake_group_id, {'status': 'error'})
            (self.manager.driver.schedule_create_share_group.
                assert_called_once_with(self.context, group_id, request_spec,
                                        {}))

    @mock.patch.object(db, 'share_group_update', mock.Mock())
    def test_create_group_exception_puts_group_in_error_state(self):
        """Test that exceptions for create_share_group.

        Puts the share in 'error' state and raises the exception.
        """

        fake_group_id = 1
        group_id = fake_group_id
        request_spec = {"share_group_id": group_id}
        with mock.patch.object(self.manager.driver,
                               'schedule_create_share_group',
                               mock.Mock(side_effect=exception.NotFound)):
            self.assertRaises(exception.NotFound,
                              self.manager.create_share_group,
                              self.context, fake_group_id,
                              request_spec=request_spec,
                              filter_properties={})

    def test_migrate_share_to_host(self):

        class fake_host(object):
            host = 'fake@backend#pool'

        share = db_utils.create_share()
        host = fake_host()

        self.mock_object(db, 'share_get', mock.Mock(return_value=share))
        self.mock_object(share_rpcapi.ShareAPI, 'migration_start',
                         mock.Mock(side_effect=TypeError))
        self.mock_object(base.Scheduler,
                         'host_passes_filters',
                         mock.Mock(return_value=host))

        self.assertRaises(
            TypeError, self.manager.migrate_share_to_host,
            self.context, share['id'], 'fake@backend#pool', False, True, True,
            False, True, 'fake_net_id', 'fake_type_id', {}, None)

        db.share_get.assert_called_once_with(self.context, share['id'])
        base.Scheduler.host_passes_filters.assert_called_once_with(
            self.context, 'fake@backend#pool', {}, None)
        share_rpcapi.ShareAPI.migration_start.assert_called_once_with(
            self.context, share, host.host, False, True, True, False, True,
            'fake_net_id', 'fake_type_id')

    @ddt.data(exception.NoValidHost(reason='fake'), TypeError)
    def test_migrate_share_to_host_exception(self, exc):

        share = db_utils.create_share(status=constants.STATUS_MIGRATING)
        host = 'fake@backend#pool'
        request_spec = {'share_id': share['id']}

        self.mock_object(db, 'share_get', mock.Mock(return_value=share))
        self.mock_object(
            base.Scheduler, 'host_passes_filters',
            mock.Mock(side_effect=exc))
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_instance_update')

        capture = (exception.NoValidHost if
                   isinstance(exc, exception.NoValidHost) else TypeError)

        self.assertRaises(
            capture, self.manager.migrate_share_to_host,
            self.context, share['id'], host, False, True, True, False, True,
            'fake_net_id', 'fake_type_id', request_spec, None)

        base.Scheduler.host_passes_filters.assert_called_once_with(
            self.context, host, request_spec, None)
        db.share_get.assert_called_once_with(self.context, share['id'])
        db.share_update.assert_called_once_with(
            self.context, share['id'],
            {'task_state': constants.TASK_STATE_MIGRATION_ERROR})
        db.share_instance_update.assert_called_once_with(
            self.context, share.instance['id'],
            {'status': constants.STATUS_AVAILABLE})

    def test_manage_share(self):

        share = db_utils.create_share()

        self.mock_object(db, 'share_get', mock.Mock(return_value=share))
        self.mock_object(share_rpcapi.ShareAPI, 'manage_share')
        self.mock_object(base.Scheduler, 'host_passes_filters')

        self.manager.manage_share(self.context, share['id'], 'driver_options',
                                  {}, None)

    def test_manage_share_exception(self):

        share = db_utils.create_share()

        db_update = self.mock_object(db, 'share_update', mock.Mock())
        self.mock_object(
            base.Scheduler, 'host_passes_filters',
            mock.Mock(side_effect=exception.NoValidHost('fake')))

        share_id = share['id']

        self.assertRaises(
            exception.NoValidHost, self.manager.manage_share,
            self.context, share['id'], 'driver_options',
            {'share_id': share_id}, None)
        db_update.assert_called_once_with(
            self.context, share_id,
            {'status': constants.STATUS_MANAGE_ERROR, 'size': 1})

    def test_create_share_replica_exception_path(self):
        """Test 'raisable' exceptions for create_share_replica."""
        db_update = self.mock_object(db, 'share_replica_update')
        self.mock_object(db, 'share_snapshot_instance_get_all_with_filters',
                         mock.Mock(return_value=[{'id': '123'}]))
        snap_update = self.mock_object(db, 'share_snapshot_instance_update')
        request_spec = fakes.fake_replica_request_spec()
        replica_id = request_spec.get('share_instance_properties').get('id')
        expected_updates = {
            'status': constants.STATUS_ERROR,
            'replica_state': constants.STATUS_ERROR,
        }
        with mock.patch.object(self.manager.driver, 'schedule_create_replica',
                               mock.Mock(side_effect=exception.NotFound)):

            self.assertRaises(exception.NotFound,
                              self.manager.create_share_replica,
                              self.context,
                              request_spec=request_spec,
                              filter_properties={})
            db_update.assert_called_once_with(
                self.context, replica_id, expected_updates)
            snap_update.assert_called_once_with(
                self.context, '123', {'status': constants.STATUS_ERROR})

    def test_create_share_replica_no_valid_host(self):
        """Test the NoValidHost exception for create_share_replica."""
        db_update = self.mock_object(db, 'share_replica_update')
        request_spec = fakes.fake_replica_request_spec()
        replica_id = request_spec.get('share_instance_properties').get('id')
        expected_updates = {
            'status': constants.STATUS_ERROR,
            'replica_state': constants.STATUS_ERROR,
        }
        with mock.patch.object(
                self.manager.driver, 'schedule_create_replica',
                mock.Mock(side_effect=self.raise_no_valid_host)):

            retval = self.manager.create_share_replica(
                self.context, request_spec=request_spec, filter_properties={})

            self.assertIsNone(retval)
            db_update.assert_called_once_with(
                self.context, replica_id, expected_updates)

    def test_create_share_replica(self):
        """Test happy path for create_share_replica."""
        db_update = self.mock_object(db, 'share_replica_update')
        mock_scheduler_driver_call = self.mock_object(
            self.manager.driver, 'schedule_create_replica')
        request_spec = fakes.fake_replica_request_spec()

        retval = self.manager.create_share_replica(
            self.context, request_spec=request_spec, filter_properties={})

        mock_scheduler_driver_call.assert_called_once_with(
            self.context, request_spec, {})
        self.assertFalse(db_update.called)
        self.assertIsNone(retval)



# Copyright (c) 2015 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
NetApp cDOT CIFS protocol helper class.
"""

import re

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share.drivers.netapp.dataontap.protocols import base
from manila.share.drivers.netapp import utils as na_utils


class NetAppCmodeCIFSHelper(base.NetAppBaseHelper):
    """NetApp cDOT CIFS protocol helper class."""

    @na_utils.trace
    def create_share(self, share, share_name,
                     clear_current_export_policy=True):
        """Creates CIFS share on Data ONTAP Vserver."""
        self._client.create_cifs_share(share_name)
        if clear_current_export_policy:
            self._client.remove_cifs_share_access(share_name, 'Everyone')

        # Ensure 'ntfs' security style
        self._client.set_volume_security_style(share_name,
                                               security_style='ntfs')

        # Return a callback that may be used for generating export paths
        # for this share.
        return (lambda export_address, share_name=share_name:
                r'\\%s\%s' % (export_address, share_name))

    @na_utils.trace
    def delete_share(self, share, share_name):
        """Deletes CIFS share on Data ONTAP Vserver."""
        host_ip, share_name = self._get_export_location(share)
        self._client.remove_cifs_share(share_name)

    @na_utils.trace
    @base.access_rules_synchronized
    def update_access(self, share, share_name, rules):
        """Replaces the list of access rules known to the backend storage."""

        # Ensure rules are valid
        for rule in rules:
            self._validate_access_rule(rule)

        new_rules = {rule['access_to']: rule['access_level'] for rule in rules}

        # Get rules from share
        existing_rules = self._get_access_rules(share, share_name)

        # Update rules in an order that will prevent transient disruptions
        self._handle_added_rules(share_name, existing_rules, new_rules)
        self._handle_ro_to_rw_rules(share_name, existing_rules, new_rules)
        self._handle_rw_to_ro_rules(share_name, existing_rules, new_rules)
        self._handle_deleted_rules(share_name, existing_rules, new_rules)

    @na_utils.trace
    def _validate_access_rule(self, rule):
        """Checks whether access rule type and level are valid."""

        if rule['access_type'] != 'user':
            msg = _("Clustered Data ONTAP supports only 'user' type for "
                    "share access rules with CIFS protocol.")
            raise exception.InvalidShareAccess(reason=msg)

        if rule['access_level'] not in constants.ACCESS_LEVELS:
            raise exception.InvalidShareAccessLevel(level=rule['access_level'])

    @na_utils.trace
    def _handle_added_rules(self, share_name, existing_rules, new_rules):
        """Updates access rules added between two rule sets."""
        added_rules = {
            user_or_group: permission
            for user_or_group, permission in new_rules.items()
            if user_or_group not in existing_rules
        }

        for user_or_group, permission in added_rules.items():
            self._client.add_cifs_share_access(
                share_name, user_or_group, self._is_readonly(permission))

    @na_utils.trace
    def _handle_ro_to_rw_rules(self, share_name, existing_rules, new_rules):
        """Updates access rules modified (RO-->RW) between two rule sets."""
        modified_rules = {
            user_or_group: permission
            for user_or_group, permission in new_rules.items()
            if (user_or_group in existing_rules and
                permission == constants.ACCESS_LEVEL_RW and
                existing_rules[user_or_group] != 'full_control')
        }

        for user_or_group, permission in modified_rules.items():
            self._client.modify_cifs_share_access(
                share_name, user_or_group, self._is_readonly(permission))

    @na_utils.trace
    def _handle_rw_to_ro_rules(self, share_name, existing_rules, new_rules):
        """Returns access rules modified (RW-->RO) between two rule sets."""
        modified_rules = {
            user_or_group: permission
            for user_or_group, permission in new_rules.items()
            if (user_or_group in existing_rules and
                permission == constants.ACCESS_LEVEL_RO and
                existing_rules[user_or_group] != 'read')
        }

        for user_or_group, permission in modified_rules.items():
            self._client.modify_cifs_share_access(
                share_name, user_or_group, self._is_readonly(permission))

    @na_utils.trace
    def _handle_deleted_rules(self, share_name, existing_rules, new_rules):
        """Returns access rules deleted between two rule sets."""
        deleted_rules = {
            user_or_group: permission
            for user_or_group, permission in existing_rules.items()
            if user_or_group not in new_rules
        }

        for user_or_group, permission in deleted_rules.items():
            self._client.remove_cifs_share_access(share_name, user_or_group)

    @na_utils.trace
    def _get_access_rules(self, share, share_name):
        """Returns the list of access rules known to the backend storage."""
        return self._client.get_cifs_share_access(share_name)

    @na_utils.trace
    def get_target(self, share):
        """Returns OnTap target IP based on share export location."""
        return self._get_export_location(share)[0]

    @na_utils.trace
    def get_share_name_for_share(self, share):
        """Returns the flexvol name that hosts a share."""
        _, share_name = self._get_export_location(share)
        return share_name

    @staticmethod
    def _get_export_location(share):
        """Returns host ip and share name for a given CIFS share."""
        export_location = share['export_location'] or '\\\\\\'
        regex = r'^(?:\\\\|//)(?P<host_ip>.*)(?:\\|/)(?P<share_name>.*)$'
        match = re.match(regex, export_location)
        if match:
            return match.group('host_ip'), match.group('share_name')
        else:
            return '', ''



# Copyright (c) 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
from webob import exc

from manila.api.openstack import api_version_request as api_version
from manila.api.v2 import share_export_locations as export_locations
from manila.common import constants
from manila import context
from manila import db
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils


@ddt.ddt
class ShareExportLocationsAPITest(test.TestCase):

    def _get_request(self, version="2.9", use_admin_context=True):
        req = fakes.HTTPRequest.blank(
            '/v2/shares/%s/export_locations' % self.share_instance_id,
            version=version, use_admin_context=use_admin_context)
        return req

    def setUp(self):
        super(ShareExportLocationsAPITest, self).setUp()
        self.controller = (
            export_locations.ShareExportLocationController())
        self.resource_name = self.controller.resource_name
        self.ctxt = {
            'admin': context.RequestContext('admin', 'fake', True),
            'user': context.RequestContext('fake', 'fake'),
        }
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.share = db_utils.create_share()
        self.share_instance_id = self.share.instance.id
        self.req = self._get_request()
        paths = ['fake1/1/', 'fake2/2', 'fake3/3']
        db.share_export_locations_update(
            self.ctxt['admin'], self.share_instance_id, paths, False)

    @ddt.data({'role': 'admin', 'version': '2.9'},
              {'role': 'user', 'version': '2.9'},
              {'role': 'admin', 'version': '2.13'},
              {'role': 'user', 'version': '2.13'})
    @ddt.unpack
    def test_list_and_show(self, role, version):

        summary_keys = ['id', 'path']
        admin_summary_keys = summary_keys + [
            'share_instance_id', 'is_admin_only']
        detail_keys = summary_keys + ['created_at', 'updated_at']
        admin_detail_keys = admin_summary_keys + ['created_at', 'updated_at']

        self._test_list_and_show(role, version, summary_keys, detail_keys,
                                 admin_summary_keys, admin_detail_keys)

    @ddt.data('admin', 'user')
    def test_list_and_show_with_preferred_flag(self, role):

        summary_keys = ['id', 'path', 'preferred']
        admin_summary_keys = summary_keys + [
            'share_instance_id', 'is_admin_only']
        detail_keys = summary_keys + ['created_at', 'updated_at']
        admin_detail_keys = admin_summary_keys + ['created_at', 'updated_at']

        self._test_list_and_show(role, '2.14', summary_keys, detail_keys,
                                 admin_summary_keys, admin_detail_keys)

    def _test_list_and_show(self, role, version, summary_keys, detail_keys,
                            admin_summary_keys, admin_detail_keys):

        req = self._get_request(version=version,
                                use_admin_context=(role == 'admin'))
        index_result = self.controller.index(req, self.share['id'])

        self.assertIn('export_locations', index_result)
        self.assertEqual(1, len(index_result))
        self.assertEqual(3, len(index_result['export_locations']))

        for index_el in index_result['export_locations']:
            self.assertIn('id', index_el)
            show_result = self.controller.show(
                req, self.share['id'], index_el['id'])
            self.assertIn('export_location', show_result)
            self.assertEqual(1, len(show_result))

            show_el = show_result['export_location']

            # Check summary keys in index result & detail keys in show result
            if role == 'admin':
                self.assertEqual(len(admin_summary_keys), len(index_el))
                for key in admin_summary_keys:
                    self.assertIn(key, index_el)
                self.assertEqual(len(admin_detail_keys), len(show_el))
                for key in admin_detail_keys:
                    self.assertIn(key, show_el)
            else:
                self.assertEqual(len(summary_keys), len(index_el))
                for key in summary_keys:
                    self.assertIn(key, index_el)
                self.assertEqual(len(detail_keys), len(show_el))
                for key in detail_keys:
                    self.assertIn(key, show_el)

            # Ensure keys common to index & show results have matching values
            for key in summary_keys:
                self.assertEqual(index_el[key], show_el[key])

    def test_list_export_locations_share_not_found(self):
        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.index,
            self.req, 'inexistent_share_id',
        )

    def test_show_export_location_share_not_found(self):
        index_result = self.controller.index(self.req, self.share['id'])
        el_id = index_result['export_locations'][0]['id']
        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.show,
            self.req, 'inexistent_share_id', el_id,
        )

    def test_show_export_location_not_found(self):
        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.show,
            self.req, self.share['id'], 'inexistent_export_location',
        )

    def test_get_admin_export_location(self):
        el_data = {
            'path': '/admin/export/location',
            'is_admin_only': True,
            'metadata': {'foo': 'bar'},
        }
        db.share_export_locations_update(
            self.ctxt['admin'], self.share_instance_id, el_data, True)
        index_result = self.controller.index(self.req, self.share['id'])
        el_id = index_result['export_locations'][0]['id']

        # Not found for member
        member_req = self._get_request(use_admin_context=False)
        self.assertRaises(
            exc.HTTPForbidden,
            self.controller.show,
            member_req, self.share['id'], el_id,
        )

        # Ok for admin
        el = self.controller.show(self.req, self.share['id'], el_id)
        for k, v in el.items():
            self.assertEqual(v, el[k])

    @ddt.data(*set(('2.46', '2.47', api_version._MAX_API_VERSION)))
    def test_list_export_locations_replicated_share(self, version):
        """Test the export locations API changes between 2.46 and 2.47

        For API version <= 2.46, non-active replica export locations are
        included in the API response. They are not included in and beyond
        version 2.47.
        """
        # Setup data
        share = db_utils.create_share(
            replication_type=constants.REPLICATION_TYPE_READABLE,
            replica_state=constants.REPLICA_STATE_ACTIVE)
        active_replica_id = share.instance.id
        exports = [
            {'path': 'myshare.mydomain/active-replica-exp1',
             'is_admin_only': False},
            {'path': 'myshare.mydomain/active-replica-exp2',
             'is_admin_only': False},
        ]
        db.share_export_locations_update(
            self.ctxt['user'], active_replica_id, exports)

        # Replicas
        share_replica2 = db_utils.create_share_replica(
            share_id=share.id, replica_state=constants.REPLICA_STATE_IN_SYNC)
        share_replica3 = db_utils.create_share_replica(
            share_id=share.id,
            replica_state=constants.REPLICA_STATE_OUT_OF_SYNC)
        replica2_exports = [
            {'path': 'myshare.mydomain/insync-replica-exp',
             'is_admin_only': False}
        ]
        replica3_exports = [
            {'path': 'myshare.mydomain/outofsync-replica-exp',
             'is_admin_only': False}
        ]
        db.share_export_locations_update(
            self.ctxt['user'], share_replica2.id, replica2_exports)
        db.share_export_locations_update(
            self.ctxt['user'], share_replica3.id, replica3_exports)

        req = self._get_request(version=version)
        index_result = self.controller.index(req, share['id'])

        actual_paths = [el['path'] for el in index_result['export_locations']]
        if self.is_microversion_ge(version, '2.47'):
            self.assertEqual(2, len(index_result['export_locations']))
            self.assertNotIn(
                'myshare.mydomain/insync-replica-exp', actual_paths)
            self.assertNotIn(
                'myshare.mydomain/outofsync-replica-exp', actual_paths)
        else:
            self.assertEqual(4, len(index_result['export_locations']))
            self.assertIn('myshare.mydomain/insync-replica-exp', actual_paths)
            self.assertIn(
                'myshare.mydomain/outofsync-replica-exp', actual_paths)

    @ddt.data('1.0', '2.0', '2.8')
    def test_list_with_unsupported_version(self, version):
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.index,
            self._get_request(version),
            self.share_instance_id,
        )

    @ddt.data('1.0', '2.0', '2.8')
    def test_show_with_unsupported_version(self, version):
        index_result = self.controller.index(self.req, self.share['id'])

        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.show,
            self._get_request(version),
            self.share['id'],
            index_result['export_locations'][0]['id']
        )



# Copyright 2015 Goutham Pacha Ravi
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common


class ReplicationViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = 'share_replicas'
    _collection_links = 'share_replica_links'

    _detail_version_modifiers = [
        "add_cast_rules_to_readonly_field",
    ]

    def summary_list(self, request, replicas):
        """Summary view of a list of replicas."""
        return self._list_view(self.summary, request, replicas)

    def detail_list(self, request, replicas):
        """Detailed view of a list of replicas."""
        return self._list_view(self.detail, request, replicas)

    def summary(self, request, replica):
        """Generic, non-detailed view of a share replica."""

        replica_dict = {
            'id': replica.get('id'),
            'share_id': replica.get('share_id'),
            'status': replica.get('status'),
            'replica_state': replica.get('replica_state'),
        }
        return {'share_replica': replica_dict}

    def detail(self, request, replica):
        """Detailed view of a single replica."""
        context = request.environ['manila.context']

        replica_dict = {
            'id': replica.get('id'),
            'share_id': replica.get('share_id'),
            'availability_zone': replica.get('availability_zone'),
            'created_at': replica.get('created_at'),
            'status': replica.get('status'),
            'share_network_id': replica.get('share_network_id'),
            'replica_state': replica.get('replica_state'),
            'updated_at': replica.get('updated_at'),
        }

        if context.is_admin:
            replica_dict['share_server_id'] = replica.get('share_server_id')
            replica_dict['host'] = replica.get('host')

        self.update_versioned_resource_dict(request, replica_dict, replica)
        return {'share_replica': replica_dict}

    def _list_view(self, func, request, replicas):
        """Provide a view for a list of replicas."""

        replicas_list = [func(request, replica)['share_replica']
                         for replica in replicas]

        replica_links = self._get_collection_links(request,
                                                   replicas,
                                                   self._collection_name)
        replicas_dict = {self._collection_name: replicas_list}

        if replica_links:
            replicas_dict[self._collection_links] = replica_links

        return replicas_dict

    @common.ViewBuilder.versioned_method("2.30")
    def add_cast_rules_to_readonly_field(self, context, replica_dict, replica):
        if context.is_admin:
            replica_dict['cast_rules_to_readonly'] = replica.get(
                'cast_rules_to_readonly', False)



# Copyright (c) 2015 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from zaqarclient.queues import client as zaqar

zaqar_notification_opts = [
    cfg.StrOpt(
        "zaqar_username",
        help="Username that should be used for init of zaqar client.",
    ),
    cfg.StrOpt(
        "zaqar_password",
        secret=True,
        help="Password for user specified in opt 'zaqar_username'.",
    ),
    cfg.StrOpt(
        "zaqar_project_name",
        help=("Project/Tenant name that is owns user specified "
              "in opt 'zaqar_username'."),
    ),
    cfg.StrOpt(
        "zaqar_auth_url",
        default="http://127.0.0.1:35357/v2.0/",
        help="Auth url to be used by Zaqar client.",
    ),
    cfg.StrOpt(
        "zaqar_region_name",
        help="Name of the region that should be used. Optional.",
    ),
    cfg.StrOpt(
        "zaqar_service_type",
        default="messaging",
        help="Service type for Zaqar. Optional.",
    ),
    cfg.StrOpt(
        "zaqar_endpoint_type",
        default="publicURL",
        help="Type of endpoint to be used for init of Zaqar client. Optional.",
    ),
    cfg.FloatOpt(
        "zaqar_api_version",
        default=1.1,
        help="Version of Zaqar API to use. Optional.",
    ),
    cfg.ListOpt(
        "zaqar_queues",
        default=["manila_notification_qeueue"],
        help=("List of queues names to be used for sending Manila "
              "notifications. Optional."),
    ),
]

CONF = cfg.CONF
CONF.register_opts(zaqar_notification_opts, group='zaqar')

ZAQARCLIENT = zaqar.Client(
    version=CONF.zaqar.zaqar_api_version,
    conf={
        "auth_opts": {
            "backend": "keystone",
            "options": {
                "os_username": CONF.zaqar.zaqar_username,
                "os_password": CONF.zaqar.zaqar_password,
                "os_project_name": CONF.zaqar.zaqar_project_name,
                "os_auth_url": CONF.zaqar.zaqar_auth_url,
                "os_region_name": CONF.zaqar.zaqar_region_name,
                "os_service_type": CONF.zaqar.zaqar_service_type,
                "os_endpoint_type": CONF.zaqar.zaqar_endpoint_type,
                "insecure": True,
            },
        },
    },
)
